# Copyright (c) OpenMMLab. All rights reserved.
# The code is modified from https://github.com/visionml/pytracking/blob/master/pytracking/analysis/plot_results.py # noqa: E501

from typing import List, Optional

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from mmengine.utils import mkdir_or_exist

PALETTE = [(1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0), (1.0, 0.0, 1.0),
           (0.0, 1.0, 1.0), (0.5, 0.5, 0.5),
           (136.0 / 255.0, 0.0, 21.0 / 255.0),
           (1.0, 127.0 / 255.0, 39.0 / 255.0),
           (0.0, 162.0 / 255.0, 232.0 / 255.0),
           (0.0, 0.5, 0.0), (1.0, 0.5, 0.2), (0.1, 0.4, 0.0), (0.6, 0.3, 0.9),
           (0.4, 0.7, 0.1), (0.2, 0.1, 0.7), (0.7, 0.6, 0.2),
           (1.0, 102.0 / 255.0, 102.0 / 255.0),
           (153.0 / 255.0, 1.0, 153.0 / 255.0),
           (102.0 / 255.0, 102.0 / 255.0, 1.0),
           (1.0, 192.0 / 255.0, 203.0 / 255.0)]
LINE_STYLE = ['-'] * len(PALETTE)


def plot_sot_curve(y: np.ndarray,
                   x: np.ndarray,
                   scores: np.ndarray,
                   tracker_names: List,
                   plot_opts: dict,
                   plot_save_path: Optional[str] = None,
                   show: bool = False):
    """Plot curves for SOT.

    Args:
        y (np.ndarray): The content along the Y axis. It has shape (N, M),
            where N is the number of trackers and M is the number of values
            corresponding to the X.
        x (np.ndarray): The content along the X axis. It has shape (M).
        scores (np.ndarray): The content of viualized indicators.
        tracker_names (List): The names of trackers.
        plot_opts (dict): The options for plot.
        plot_save_path (Optional[str], optional): The saved path of the figure.
            Defaults to None.
        show (bool, optional): Whether to show. Defaults to False.
    """
    x, scores = x.squeeze(), scores.squeeze()
    assert y.ndim == 2 and x.ndim == 1 and scores.ndim == 1

    # Plot settings
    font_size = plot_opts.get('font_size', 12)
    font_size_axis = plot_opts.get('font_size_axis', 13)
    line_width = plot_opts.get('line_width', 2)
    font_size_legend = plot_opts.get('font_size_legend', 13)

    plot_type = plot_opts['plot_type']
    legend_loc = plot_opts['legend_loc']

    xlabel = plot_opts['xlabel']
    ylabel = plot_opts['ylabel']
    xlim = plot_opts['xlim']
    ylim = plot_opts['ylim']

    title = plot_opts['title']

    matplotlib.rcParams.update({'font.size': font_size})
    matplotlib.rcParams.update({'axes.titlesize': font_size_axis})
    matplotlib.rcParams.update({'axes.titleweight': 'black'})
    matplotlib.rcParams.update({'axes.labelsize': font_size_axis})

    # Plot curves
    fig, ax = plt.subplots()

    index_sort = np.argsort(scores)
    plotted_lines = []
    legend_text = []

    for id, id_sort in enumerate(index_sort):
        line = ax.plot(
            x.tolist(),
            y[id_sort, :].tolist(),
            linewidth=line_width,
            color=PALETTE[len(index_sort) - id - 1],
            linestyle=LINE_STYLE[len(index_sort) - id - 1])

        plotted_lines.append(line[0])
        legend_text.append('{} [{:.1f}]'.format(tracker_names[id_sort],
                                                scores[id_sort]))

    ax.legend(
        plotted_lines[::-1],
        legend_text[::-1],
        loc=legend_loc,
        fancybox=False,
        edgecolor='black',
        fontsize=font_size_legend,
        framealpha=1.0)
    ax.set(xlabel=xlabel, ylabel=ylabel, xlim=xlim, ylim=ylim, title=title)
    ax.grid(True, linestyle='-.')
    fig.tight_layout()

    if plot_save_path is not None:
        mkdir_or_exist(plot_save_path)
        fig.savefig(
            '{}/{}_plot.pdf'.format(plot_save_path, plot_type),
            dpi=300,
            format='pdf',
            transparent=True)
    plt.draw()
    if show:
        plt.show()


def plot_success_curve(success: np.ndarray,
                       tracker_names: List,
                       plot_opts: Optional[dict] = None,
                       plot_save_path: Optional[str] = None,
                       show: bool = False):
    """Plot curves of Success for SOT.

    Args:
        success (np.ndarray): The content of viualized indicators. It has shape
            (N, M), where N is the number of trackers and M is the number of
            ``Success`` corresponding to the X.
        tracker_names (List): The names of trackers.
        plot_opts (Optional[dict], optional): The options for plot.
            Defaults to None.
        plot_save_path (Optional[str], optional): The saved path of the figure.
            Defaults to None.
        show (bool, optional): Whether to show. Defaults to False.
    """
    assert len(tracker_names) == len(success)
    success_plot_opts = {
        'plot_type': 'success',
        'legend_loc': 'lower left',
        'xlabel': 'Overlap threshold',
        'ylabel': 'Overlap Precision [%]',
        'xlim': (0, 1.0),
        'ylim': (0, 100),
        'title': 'Success plot'
    }
    if plot_opts is not None:
        success_plot_opts.update(success_plot_opts)
    success_scores = np.mean(success, axis=1)

    plot_sot_curve(success, np.arange(0, 1.05, 0.05), success_scores,
                   tracker_names, success_plot_opts, plot_save_path, show)


def plot_norm_precision_curve(norm_precision: np.ndarray,
                              tracker_names: List,
                              plot_opts: Optional[dict] = None,
                              plot_save_path: Optional[str] = None,
                              show: bool = False):
    """Plot curves of Norm Precision for SOT.

    Args:
        norm_precision (np.ndarray): The content of viualized indicators. It
            has shape (N, M), where N is the number of trackers and M is the
            number of ``Norm Precision`` corresponding to the X.
        tracker_names (List): The names of trackers.
        plot_opts (Optional[dict], optional): The options for plot.
            Defaults to None.
        plot_save_path (Optional[str], optional): The saved path of the figure.
            Defaults to None.
        show (bool, optional): Whether to show. Defaults to False.
    """
    assert len(tracker_names) == len(norm_precision)
    norm_precision_plot_opts = {
        'plot_type': 'norm_precision',
        'legend_loc': 'lower right',
        'xlabel': 'Location error threshold',
        'ylabel': 'Distance Precision [%]',
        'xlim': (0, 0.5),
        'ylim': (0, 100),
        'title': 'Normalized Precision plot'
    }
    if plot_opts is not None:
        norm_precision_plot_opts.update(norm_precision_plot_opts)

    plot_sot_curve(norm_precision, np.arange(0, 0.51, 0.01),
                   norm_precision[:, 20], tracker_names,
                   norm_precision_plot_opts, plot_save_path, show)


def plot_precision_curve(precision: np.ndarray,
                         tracker_names: List,
                         plot_opts: Optional[dict] = None,
                         plot_save_path: Optional[str] = None,
                         show: bool = False):
    """Plot curves of Precision for SOT.

    Args:
        precision (np.ndarray): The content of viualized indicators. It has
            shape (N, M), where N is the number of trackers and M is the
            number of ``Precision`` corresponding to the X.
        tracker_names (List): The names of trackers.
        plot_opts (Optional[dict], optional): The options for plot.
            Defaults to None.
        plot_save_path (Optional[str], optional): The saved path of the figure.
            Defaults to None.
        show (bool, optional): Whether to show. Defaults to False.
    """
    assert len(tracker_names) == len(precision)
    precision_plot_opts = {
        'plot_type': 'precision',
        'legend_loc': 'lower right',
        'xlabel': 'Location error threshold [pixels]',
        'ylabel': 'Distance Precision [%]',
        'xlim': (0, 50),
        'ylim': (0, 100),
        'title': 'Precision plot'
    }
    if plot_opts is not None:
        precision_plot_opts.update(plot_opts)

    plot_sot_curve(precision, np.arange(0, 51, 1), precision[:, 20],
                   tracker_names, precision_plot_opts, plot_save_path, show)



# Copyright (c) OpenMMLab. All rights reserved.
import os
import os.path as osp
import shutil
import tempfile
from collections import OrderedDict
from copy import deepcopy
from typing import Dict, List, Optional, Sequence, Union

import mmengine
import numpy as np
from mmengine.logging import MMLogger
from mmengine.utils import mkdir_or_exist
from tabulate import tabulate

from mmtrack.registry import METRICS
from mmtrack.utils import format_video_level_show
from ..functional import (eval_sot_accuracy_robustness, eval_sot_eao,
                          eval_sot_ope)
from .base_video_metrics import BaseVideoMetric


@METRICS.register_module()
class SOTMetric(BaseVideoMetric):
    """SOT evaluation metrics.

    Args:
        metric (Union[str, Sequence[str]], optional): Metrics to be evaluated.
            Valid metrics are included in ``self.allowed_metrics``.
            Defaults to 'OPE'.
        metric_options (Optional[dict], optional): Options for calculating
            metrics. Defaults to dict(dataset_type='vot2018',
            only_eval_visible=False).
        format_only (bool, optional): If True, only formatting the results to
            the official format and not performing evaluation.
            Defaults to False.
        outfile_prefix (Optional[str], optional): The prefix of json files. It
            includes the file path and the prefix of filename,
            e.g., "a/b/prefix". If not specified, a temp file will be created.
            Defaults to None.
        collect_device (str, optional): Device name used for collecting results
            from different ranks during distributed training. Must be 'cpu' or
            'gpu'. Defaults to 'cpu'.
        prefix (Optional[str], optional): The prefix that will be added in the
            metric names to disambiguate homonymous metrics of different
            evaluators. If prefix is not provided in the argument,
            self.default_prefix will be used instead. Defaults to None.
        saved_track_res_path (Optional[str], optional): The saved path of
            tracked results. Defaults to None.
        options_after_eval (Optional[dict], optional): The options after
            evaluation. Defaults to {}. The usage is the following:
            ```
                options_after_eval = dict(
                    saved_eval_res_file = './results/sot_res.json',
                    tracker_name = 'siamrpn++',
                    eval_show_video_indices = 10)
            ```
            Here, ``eval_show_video_indices`` is used to index a numpy.ndarray.
            It can be int (positive or negative) or list.
            ``saved_eval_res_file`` must be a json/yaml/pickle file.
    """
    default_prefix: Optional[str] = 'sot'
    allowed_metrics = ['OPE', 'VOT']
    allowed_metric_options = ['dataset_type', 'only_eval_visible']
    VOT_INTERVAL = dict(vot2018=[100, 356], vot2019=[46, 291])

    def __init__(self,
                 metric: Union[str, Sequence[str]] = 'OPE',
                 metric_options: Optional[dict] = dict(
                     dataset_type='vot2018', only_eval_visible=False),
                 format_only: bool = False,
                 outfile_prefix: Optional[str] = None,
                 collect_device: str = 'cpu',
                 prefix: Optional[str] = None,
                 saved_track_res_path: Optional[str] = None,
                 options_after_eval: dict = {}) -> None:
        super().__init__(collect_device=collect_device, prefix=prefix)
        self.metrics = metric if isinstance(metric, list) else [metric]
        assert not (
            'OPE' in self.metrics and 'VOT' in self.metrics
        ), 'We can not evaluate one tracking result on both OPE and '
        'VOT metrics since the track result on VOT mode '
        'may be not the true bbox coordinates.'
        self.metric_options = metric_options
        for metric in self.metrics:
            if metric not in self.allowed_metrics:
                raise KeyError(
                    f'metric should be in {str(self.allowed_metrics)}, '
                    f'but got {metric}.')
        for metric_option in self.metric_options:
            if metric_option not in self.allowed_metric_options:
                raise KeyError(
                    f'metric option should be in {str(self.allowed_metric_options)}, '  # noqa: E501
                    f'but got {metric_option}.')
        self.outfile_prefix = outfile_prefix
        self.format_only = format_only
        self.saved_track_res_path = saved_track_res_path
        self.options_after_eval = options_after_eval
        self.preds_per_video, self.gts_per_video = [], []
        self.frame_ids, self.visible_per_video = [], []

    def process(self, data_batch: dict, data_samples: Sequence[dict]) -> None:
        """Process one batch of data samples and predictions. The processed
        results should be stored in ``self.results``, which will be used to
        compute the metrics when all batches have been processed.

        Args:
            data_batch (dict): A batch of data from the dataloader.
            data_samples (Sequence[dict]): A batch of data samples that
                contain annotations and predictions.
        """
        for data_sample in data_samples:
            data_instance = data_sample['instances'][0]

            self.preds_per_video.append(
                data_sample['pred_track_instances']['bboxes'][0].cpu().numpy())
            if 'bbox' in data_instance:
                self.gts_per_video.append(data_instance['bbox'])
            else:
                assert self.format_only, 'If there is no ground truth '
                "bounding bbox, 'format_only' must be True"
            self.visible_per_video.append(data_instance['visible'])
            self.frame_ids.append(data_sample['frame_id'])

            if data_sample['frame_id'] == data_sample['video_length'] - 1:
                img_path_split = data_sample['img_path'].split(os.sep)
                # The ``img_path`` in LaSOT, OTB100 and VOT2018 have an extra
                # common directory outside the *.jpg file.
                video_name = img_path_split[-2] if img_path_split[-2] not in [
                    'img', 'color'
                ] else img_path_split[-3]
                result = dict(
                    video_name=video_name,
                    video_id=data_sample['video_id'],
                    video_size=(data_sample['ori_shape'][1],
                                data_sample['ori_shape'][0]),
                    frame_ids=deepcopy(self.frame_ids),
                    # Collect the annotations and predictions of this video.
                    # We don't convert the ``preds_per_video`` to
                    # ``np.ndarray`` since the track results in SOT may not the
                    # tracking box in EAO metrics.
                    pred_bboxes=deepcopy(self.preds_per_video),
                    gt_bboxes=np.array(self.gts_per_video, dtype=np.float32),
                    visible=np.array(self.visible_per_video, dtype=bool))

                self.frame_ids.clear()
                self.preds_per_video.clear()
                self.gts_per_video.clear()
                self.visible_per_video.clear()

                self.results.append(result)
                break

    def compute_metrics(self, results: List) -> Dict[str, float]:
        """Compute the metrics from processed results.

        Args:
            results (List): The processed results of all data. The elements of
                the list are the processed results of one video.

        Returns:
            Dict[str, float]: The computed metrics. The keys are the names of
            the metrics, and the values are corresponding results.
        """
        logger: MMLogger = MMLogger.get_current_instance()

        # 1. Convert the format of reuslts for evaluation.
        all_pred_bboxes, all_gt_bboxes = [], []
        all_video_names, all_video_sizes, all_visible = [], [], []
        for result in results:
            all_video_names.append(result['video_name'])
            all_video_sizes.append(result['video_size'])
            all_pred_bboxes.append(result['pred_bboxes'])
            all_gt_bboxes.append(result['gt_bboxes'])
            all_visible.append(result['visible'])

        if self.saved_track_res_path is not None:
            self.save_tracked_results(all_pred_bboxes, all_video_names,
                                      self.saved_track_res_path)

        # 2. Fromat-only (Optional)
        if self.format_only:
            self.save_formatted_results(all_pred_bboxes, all_video_names)
            return dict()

        # 3. Evaluation (Optional)
        eval_results = OrderedDict()
        for metric in self.metrics:
            logger.info(f'Evaluating {metric}...')

            if metric == 'OPE':
                if self.metric_options.get('only_eval_visible', False):
                    results_ope = eval_sot_ope(all_pred_bboxes, all_gt_bboxes,
                                               all_visible)
                else:
                    results_ope = eval_sot_ope(all_pred_bboxes, all_gt_bboxes)

                ori_success = results_ope.pop('ori_success')
                ori_norm_precision = results_ope.pop('ori_norm_precision')
                ori_precision = results_ope.pop('ori_precision')
                eval_results.update(results_ope)

                saved_file_path = self.options_after_eval.get(
                    'saved_eval_res_file', None)
                if saved_file_path is not None:
                    if not saved_file_path.endswith(
                        ('.json', '.yml', '.yaml', '.pkl')):  # noqa: E125
                        raise TypeError(
                            f'Unsupported file format: {saved_file_path}. '
                            'Please specify a json, yaml or pickle file.')
                    if 'tracker_name' not in self.options_after_eval:
                        logger.warning(
                            'Not specify tracker name in the '
                            'argument options_after_eval and use the default '
                            "tracker name: 'anonymous_tracker'")
                        tracker_name = 'anonymous_tracker'
                    else:
                        tracker_name = self.options_after_eval['tracker_name']
                    mkdir_or_exist(osp.dirname(saved_file_path))
                    ori_eval_res = {
                        tracker_name:
                        dict(
                            success=np.mean(ori_success, axis=0),
                            norm_precision=np.mean(ori_norm_precision, axis=0),
                            precision=np.mean(ori_precision, axis=0))
                    }
                    mmengine.dump(ori_eval_res, saved_file_path)
                    logger.info(
                        'save evaluation results with different thresholds in '
                        f"'{saved_file_path}'")

                if self.options_after_eval.get('eval_show_video_indices',
                                               None) is not None:
                    success_per_video = np.mean(ori_success, axis=1)
                    norm_precision_per_video = ori_norm_precision[:, 20]
                    precision_per_video = ori_precision[:, 20]

                    eval_show_results = format_video_level_show(
                        all_video_names, [
                            success_per_video, norm_precision_per_video,
                            precision_per_video
                        ],
                        sort_by_first_metric=True,
                        show_indices=self.
                        options_after_eval['eval_show_video_indices'])

                    logger.info('\n' + tabulate(
                        eval_show_results,
                        headers=[
                            'video_name', 'success', 'norm_precision',
                            'precision'
                        ]))

            elif metric == 'VOT':
                if 'interval' in self.metric_options:
                    interval = self.metric_options['interval']
                else:
                    interval = self.VOT_INTERVAL.get(
                        self.metric_options['dataset_type'], None)
                eao_scores = eval_sot_eao(
                    all_pred_bboxes,
                    all_gt_bboxes,
                    videos_wh=all_video_sizes,
                    interval=interval)
                eval_results.update(eao_scores)
                accuracy_robustness = eval_sot_accuracy_robustness(
                    all_pred_bboxes, all_gt_bboxes, videos_wh=all_video_sizes)
                eval_results.update(accuracy_robustness)

            else:
                raise KeyError(
                    f"metric '{metric}' is not supported. Please use the "
                    f'metric in {str(self.allowed_metrics)}')

        return eval_results

    def save_formatted_results_got10k(self, results: List[List[np.ndarray]],
                                      video_names: List[str],
                                      outfile_prefix: str):
        """Save the formatted results in TrackingNet dataset for evaluation on
        the test server.

        Args:
            results (List[List[np.ndarray]]): The formatted results.
            video_names (List[str]): The video names.
            outfile_prefix (str): The prefix of output files.
        """
        for result, video_name in zip(results, video_names):
            video_outfile_dir = osp.join(outfile_prefix, video_name)
            if not osp.isdir(video_outfile_dir):
                os.makedirs(video_outfile_dir, exist_ok=True)
            video_bbox_txt = osp.join(video_outfile_dir,
                                      '{}_001.txt'.format(video_name))
            video_time_txt = osp.join(video_outfile_dir,
                                      '{}_time.txt'.format(video_name))
            with open(video_bbox_txt,
                      'w') as f_bbox, open(video_time_txt, 'w') as f_time:

                for bbox in result:
                    bbox = [
                        str(f'{bbox[0]:.4f}'),
                        str(f'{bbox[1]:.4f}'),
                        str(f'{(bbox[2] - bbox[0]):.4f}'),
                        str(f'{(bbox[3] - bbox[1]):.4f}')
                    ]
                    line = ','.join(bbox) + '\n'
                    f_bbox.writelines(line)
                    # We don't record testing time, so we set a default
                    # time in order to test on the server.
                    f_time.writelines('0.0001\n')

    def save_formatted_results_trackingnet(self,
                                           results: List[List[np.ndarray]],
                                           video_names: List[str],
                                           outfile_prefix: str):
        """Save the formatted results in TrackingNet dataset for evaluation on
        the test server.

        Args:
            results (List[List[np.ndarray]]): The formatted results.
            video_names (List[str]): The video names.
            outfile_prefix (str): The prefix of output files.
        """
        for result, video_name in zip(results, video_names):
            video_txt = osp.join(outfile_prefix, f'{video_name}.txt')
            with open(video_txt, 'w') as f:
                for bbox in result:
                    bbox = [
                        str(f'{bbox[0]:.4f}'),
                        str(f'{bbox[1]:.4f}'),
                        str(f'{(bbox[2] - bbox[0]):.4f}'),
                        str(f'{(bbox[3] - bbox[1]):.4f}')
                    ]
                    line = ','.join(bbox) + '\n'
                    f.writelines(line)

    def save_formatted_results(self, results: List[List[np.ndarray]],
                               video_names: List[str]):
        """Save the formatted results for evaluation on the test server.

        Args:
            results (List[List[np.ndarray]]): The formatted results.
            video_names (List[str]): The video names.
        """
        logger: MMLogger = MMLogger.get_current_instance()

        # prepare saved dir
        if self.outfile_prefix is None:
            tmp_dir = tempfile.TemporaryDirectory()
            outfile_prefix = osp.join(tmp_dir.name, 'results')
        else:
            outfile_prefix = self.outfile_prefix

        if not osp.isdir(outfile_prefix):
            os.makedirs(outfile_prefix, exist_ok=True)

        dataset_type = self.metric_options.get('dataset_type', 'got10k')
        if dataset_type == 'got10k':
            self.save_formatted_results_got10k(results, video_names,
                                               outfile_prefix)
        elif dataset_type == 'trackingnet':
            self.save_formatted_results_trackingnet(results, video_names,
                                                    outfile_prefix)
        shutil.make_archive(outfile_prefix, 'zip', outfile_prefix)
        shutil.rmtree(outfile_prefix)
        logger.info(
            f'-------- The formatted results are stored in {outfile_prefix}.zip --------'  # noqa: E501
        )

    def save_tracked_results(self, results: List[List[np.ndarray]],
                             video_names: List[str], saved_path: str):
        """Save the tracked results.

        Args:
            results (List[List[np.ndarray]]): The tracked results.
            video_names (List[str]): The video names.
            saved_path (str): The saved path of tracked results.
        """
        logger: MMLogger = MMLogger.get_current_instance()

        if not osp.isdir(saved_path):
            os.makedirs(saved_path, exist_ok=True)

        self.save_formatted_results_trackingnet(results, video_names,
                                                saved_path)
        logger.info(
            f'-------- The whole tracked results are stored in the fold {saved_path}--------'  # noqa: E501
        )



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor

from ..structures import TrackDataSample


def stack_batch(tensors: List[torch.Tensor],
                pad_size_divisor: int = 0,
                pad_value: Union[int, float] = 0) -> torch.Tensor:
    """Stack multiple tensors to form a batch and pad the images to the max
    shape use the right bottom padding mode in these images. If
    ``pad_size_divisor > 0``, add padding to ensure the common height and width
    is divisible by ``pad_size_divisor``.

    Args:
        tensors (List[Tensor]): The input multiple tensors. each is a
            TCHW 4D-tensor. T denotes the number of key/reference frames.
        pad_size_divisor (int): If ``pad_size_divisor > 0``, add padding
            to ensure the common height and width is divisible by
            ``pad_size_divisor``. This depends on the model, and many
            models need a divisibility of 32. Defaults to 0
        pad_value (int, float): The padding value. Defaults to 0

    Returns:
       Tensor: The NTCHW 5D-tensor. N denotes the batch size.
    """
    assert isinstance(tensors, list), \
        f'Expected input type to be list, but got {type(tensors)}'
    assert len(set([tensor.ndim for tensor in tensors])) == 1, \
        f'Expected the dimensions of all tensors must be the same, ' \
        f'but got {[tensor.ndim for tensor in tensors]}'
    assert tensors[0].ndim == 4, f'Expected tensor dimension to be 4, ' \
                                 f'but got {tensors[0].ndim}'
    assert len(set([tensor.shape[0] for tensor in tensors])) == 1, \
        f'Expected the channels of all tensors must be the same, ' \
        f'but got {[tensor.shape[0] for tensor in tensors]}'

    tensor_sizes = [(tensor.shape[-2], tensor.shape[-1]) for tensor in tensors]
    max_size = np.stack(tensor_sizes).max(0)

    if pad_size_divisor > 1:
        # the last two dims are H,W, both subject to divisibility requirement
        max_size = (
            max_size +
            (pad_size_divisor - 1)) // pad_size_divisor * pad_size_divisor

    padded_samples = []
    for tensor in tensors:
        padding_size = [
            0, max_size[-1] - tensor.shape[-1], 0,
            max_size[-2] - tensor.shape[-2]
        ]
        if sum(padding_size) == 0:
            padded_samples.append(tensor)
        else:
            padded_samples.append(F.pad(tensor, padding_size, value=pad_value))

    return torch.stack(padded_samples, dim=0)


def convert_data_sample_type(
        data_sample: TrackDataSample,
        num_ref_imgs: int = 1) -> Tuple[List[TrackDataSample], List[dict]]:
    """Convert the type of ``data_sample`` from dict[list] to list[dict].

    Note: This function is mainly used to be compatible with the
        interface of MMDetection. It make sure that the information of
        each reference image can be independently packed into
        ``data_sample`` in which all the keys are without prefix "ref_".

    Args:
        data_sample (TrackDataSample): Data sample input.
        num_ref_imgs (int, optional): The numbe of reference images in the
            ``data_sample``. Defaults to 1.

    Returns:
        Tuple[List[TrackDataSample], List[dict]]: The first element is the
            list of object of TrackDataSample. The second element is the
            list of meta information of reference images.
    """
    ref_data_samples, ref_metainfos = [], []
    for _ in range(num_ref_imgs):
        ref_data_samples.append(deepcopy(data_sample))
        ref_metainfos.append(deepcopy(data_sample.metainfo))

    for key, value in data_sample.metainfo.items():
        if key.startswith('ref_'):
            new_key = key[4:]
            if num_ref_imgs == 1:
                value = [value]
            assert len(value) == num_ref_imgs
            for i, v in enumerate(value):
                ref_metainfos[i][new_key] = v
                ref_data_samples[i].set_metainfo(dict(new_key=v))
                # pop the redundant original reference key.
                ref_metainfos[i].pop(key)
                ref_data_samples[i].pop(key)

    return ref_data_samples, ref_metainfos


def max_last2d(input: Tensor) -> Tuple[Tensor, Tensor]:
    """Computes the value and position of maximum in the last two dimensions.

    Args:
        input (Tensor): of shape (..., H, W)

    Returns:
        max_val (Tensor): The maximum value.
        argmax (Tensor): The position of maximum in [row, col] format.
    """

    max_val_row, argmax_row = torch.max(input, dim=-2)
    max_val, argmax_col = torch.max(max_val_row, dim=-1)
    argmax_row = argmax_row.view(argmax_col.numel(),
                                 -1)[torch.arange(argmax_col.numel()),
                                     argmax_col.view(-1)]
    argmax_row = argmax_row.reshape(argmax_col.shape)
    argmax = torch.cat((argmax_row.unsqueeze(-1), argmax_col.unsqueeze(-1)),
                       -1)
    return max_val, argmax


def format_video_level_show(
        video_names: List,
        eval_results: List[np.ndarray],
        sort_by_first_metric: bool = True,
        show_indices: Optional[Tuple[int, List]] = None) -> List[List]:
    """Format video-level performance show.

    Args:
        video_names (List): The names of the videos.
        eval_results (List[np.ndarray]): The evaluation results.
        sort_by_first_metric (bool, optional): Whether to sort the results by
            the first metric. Defaults to True.
        show_indices (Optional[Tuple[int, List]], optional): The video indices
            to be shown. Defaults to None, i.e., all videos.

    Returns:
        List[List]: The formatted video-level evaluation results. For example:
            [[`video-2`, 48.2, 49.2, 51.9],
             [`video-1`, 46.2, 48.2, 50.2]]
    """
    all_video_names_str = np.array(video_names, dtype=str)
    eval_show_results = eval_results

    if sort_by_first_metric:
        # sort from largest to smallest
        sorted_index = np.argsort(-eval_results[0])
        all_video_names_str = all_video_names_str[sorted_index]
        sorted_eval_results = []
        for eval_res in eval_results:
            sorted_eval_results.append(eval_res[sorted_index])
        eval_show_results = np.stack(sorted_eval_results).T

    if show_indices is not None:
        if isinstance(show_indices, int):
            if show_indices < 0:
                show_indices = np.arange(show_indices, 0)
            else:
                show_indices = np.arange(show_indices)
        elif isinstance(show_indices, Sequence):
            show_indices = np.array(show_indices, dtype=np.int64)
        else:
            raise NotImplementedError(
                f'{type(show_indices)} is not supported. '
                'Please use type of int or list')
        eval_show_results = eval_show_results[show_indices, :]

    eval_show_results = eval_show_results.tolist()
    for res_line, video_name in zip(eval_show_results, all_video_names_str):
        res_line.insert(0, video_name)

    return eval_show_results



_base_ = ['../_base_/default_runtime.py']

data_root = 'data/MOT17/'

img_scale = (800, 1440)
batch_size = 4

model = dict(
    data_preprocessor=dict(
        type='DetDataPreprocessor',
        pad_size_divisor=32,
        batch_augments=[
            dict(
                type='BatchSyncRandomResize',
                random_size_range=(480, 800),
                size_divisor=32,
                interval=10)
        ]),
    _scope_='mmdet',
    type='YOLOX',
    backbone=dict(type='CSPDarknet', deepen_factor=1.33, widen_factor=1.25),
    neck=dict(
        type='YOLOXPAFPN',
        in_channels=[320, 640, 1280],
        out_channels=320,
        num_csp_blocks=4),
    bbox_head=dict(
        type='YOLOXHead', num_classes=1, in_channels=320, feat_channels=320),
    train_cfg=dict(assigner=dict(type='SimOTAAssigner', center_radius=2.5)),
    test_cfg=dict(score_thr=0.01, nms=dict(type='nms', iou_threshold=0.7)),
    init_cfg=dict(
        type='Pretrained',
        checkpoint=  # noqa: E251
        'https://download.openmmlab.com/mmdetection/v2.0/yolox/yolox_x_8x8_300e_coco/yolox_x_8x8_300e_coco_20211126_140254-1ef88d67.pth'  # noqa: E501
    ))

train_pipeline = [
    dict(
        type='Mosaic',
        img_scale=img_scale,
        pad_val=114.0,
        bbox_clip_border=False),
    dict(
        type='RandomAffine',
        scaling_ratio_range=(0.1, 2),
        border=(-img_scale[0] // 2, -img_scale[1] // 2),
        bbox_clip_border=False),
    dict(
        type='MixUp',
        img_scale=img_scale,
        ratio_range=(0.8, 1.6),
        pad_val=114.0,
        bbox_clip_border=False),
    dict(type='YOLOXHSVRandomAug'),
    dict(type='RandomFlip', prob=0.5),
    dict(
        type='Resize',
        scale=img_scale,
        keep_ratio=True,
        clip_object_border=False),
    dict(type='Pad', size_divisor=32, pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(type='FilterAnnotations', min_gt_bbox_wh=(1, 1), keep_empty=False),
    dict(type='PackDetInputs')
]
test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='Resize', scale=img_scale, keep_ratio=True),
    dict(type='Pad', size_divisor=32, pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(
        type='PackDetInputs',
        meta_keys=('img_id', 'img_path', 'ori_shape', 'img_shape',
                   'scale_factor'))
]

train_dataloader = dict(
    batch_size=batch_size,
    num_workers=4,
    persistent_workers=True,
    sampler=dict(type='DefaultSampler', shuffle=True),
    dataset=dict(
        type='MultiImageMixDataset',
        _scope_='mmdet',
        dataset=dict(
            type='ConcatDataset',
            _scope_='mmdet',
            datasets=[
                dict(
                    type='CocoDataset',
                    data_root='data/MOT17',
                    ann_file='annotations/half-train_cocoformat.json',
                    data_prefix=dict(img='train'),
                    filter_cfg=dict(filter_empty_gt=True, min_size=32),
                    metainfo=dict(CLASSES=('pedestrian', )),
                    pipeline=[
                        dict(type='LoadImageFromFile'),
                        dict(type='LoadAnnotations'),
                    ]),
                dict(
                    type='CocoDataset',
                    data_root='data/crowdhuman',
                    ann_file='annotations/crowdhuman_train.json',
                    data_prefix=dict(img='train'),
                    filter_cfg=dict(filter_empty_gt=True, min_size=32),
                    metainfo=dict(CLASSES=('pedestrian', )),
                    pipeline=[
                        dict(type='LoadImageFromFile'),
                        dict(type='LoadAnnotations'),
                    ]),
                dict(
                    type='CocoDataset',
                    data_root='data/crowdhuman',
                    ann_file='annotations/crowdhuman_val.json',
                    data_prefix=dict(img='val'),
                    filter_cfg=dict(filter_empty_gt=True, min_size=32),
                    metainfo=dict(CLASSES=('pedestrian', )),
                    pipeline=[
                        dict(type='LoadImageFromFile'),
                        dict(type='LoadAnnotations'),
                    ]),
            ]),
        pipeline=train_pipeline))
val_dataloader = dict(
    batch_size=1,
    num_workers=2,
    persistent_workers=True,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='CocoDataset',
        data_root=data_root,
        _scope_='mmdet',
        ann_file='annotations/half-val_cocoformat.json',
        data_prefix=dict(img='train'),
        metainfo=dict(CLASSES=('pedestrian', )),
        test_mode=True,
        pipeline=test_pipeline))
test_dataloader = val_dataloader

# optimizer
# default 8 gpu
lr = 0.001 / 8 * batch_size
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(
        type='SGD', lr=lr, momentum=0.9, weight_decay=5e-4, nesterov=True),
    paramwise_cfg=dict(norm_decay_mult=0., bias_decay_mult=0.))

# some hyper parameters
# training settings
total_epochs = 80
num_last_epochs = 10
resume_from = None
interval = 5

train_cfg = dict(
    type='EpochBasedTrainLoop', max_epochs=total_epochs, val_interval=1)
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')
# learning policy
param_scheduler = [
    dict(
        # use quadratic formula to warm up 1 epochs
        # and lr is updated by iteration
        type='mmdet.QuadraticWarmupLR',
        by_epoch=True,
        begin=0,
        end=1,
        convert_to_iter_based=True),
    dict(
        # use cosine lr from 1 to 70 epoch
        type='mmdet.CosineAnnealingLR',
        eta_min=lr * 0.05,
        begin=1,
        T_max=total_epochs - num_last_epochs,
        end=total_epochs - num_last_epochs,
        by_epoch=True,
        convert_to_iter_based=True),
    dict(
        # use fixed lr during last 10 epochs
        type='mmdet.ConstantLR',
        by_epoch=True,
        factor=1,
        begin=total_epochs - num_last_epochs,
        end=total_epochs,
    )
]

custom_hooks = [
    dict(
        type='YOLOXModeSwitchHook',
        num_last_epochs=num_last_epochs,
        priority=48),
    dict(type='mmdet.SyncNormHook', priority=48),
    dict(
        type='mmdet.EMAHook',
        ema_type='mmdet.ExpMomentumEMA',
        momentum=0.0001,
        update_buffers=True,
        priority=49)
]
default_hooks = dict(checkpoint=dict(interval=1))

# evaluator
val_evaluator = dict(
    type='mmdet.CocoMetric',
    ann_file=data_root + 'annotations/half-val_cocoformat.json',
    metric='bbox',
    format_only=False)
test_evaluator = val_evaluator



# dataset settings
dataset_type = 'ImagenetVIDDataset'
data_root = 'data/ILSVRC/'

# data pipeline
train_pipeline = [
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[
            dict(type='LoadImageFromFile'),
            dict(type='LoadTrackAnnotations', with_instance_id=True),
            dict(type='mmdet.Resize', scale=(1000, 600), keep_ratio=True),
            dict(type='RandomFlip', prob=0.5),
        ]),
    dict(type='PackTrackInputs')
]
test_pipeline = [
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[
            dict(type='LoadImageFromFile'),
            dict(type='mmdet.Resize', scale=(1000, 600), keep_ratio=True),
        ]),
    dict(type='PackTrackInputs', pack_single_img=False)
]

# dataloader
train_dataloader = dict(
    batch_size=1,
    num_workers=2,
    persistent_workers=True,
    sampler=dict(type='DefaultSampler', shuffle=True),
    batch_sampler=dict(type='mmdet.AspectRatioBatchSampler'),
    dataset=dict(
        type='ConcatDataset',
        datasets=[
            dict(
                type=dataset_type,
                data_root=data_root,
                ann_file='annotations/imagenet_vid_train.json',
                data_prefix=dict(img_path='Data/VID'),
                filter_cfg=dict(filter_empty_gt=True, min_size=32),
                pipeline=train_pipeline,
                load_as_video=True,
                ref_img_sampler=dict(
                    num_ref_imgs=2,
                    frame_range=9,
                    filter_key_img=True,
                    method='bilateral_uniform')),
            dict(
                type=dataset_type,
                data_root=data_root,
                ann_file='annotations/imagenet_det_30plus1cls.json',
                data_prefix=dict(img_path='Data/DET'),
                filter_cfg=dict(filter_empty_gt=True, min_size=32),
                pipeline=train_pipeline,
                load_as_video=False,
                ref_img_sampler=dict(
                    num_ref_imgs=2,
                    frame_range=0,
                    filter_key_img=False,
                    method='bilateral_uniform'))
        ]))
val_dataloader = dict(
    batch_size=1,
    num_workers=2,
    persistent_workers=True,
    drop_last=False,
    sampler=dict(type='VideoSampler'),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        ann_file='annotations/imagenet_vid_val.json',
        data_prefix=dict(img_path='Data/VID'),
        pipeline=test_pipeline,
        load_as_video=True,
        ref_img_sampler=dict(
            num_ref_imgs=30,
            frame_range=[-15, 15],
            stride=1,
            method='test_with_fix_stride'),
        test_mode=True))
test_dataloader = val_dataloader

# evaluator
val_evaluator = dict(
    type='CocoVideoMetric',
    ann_file=data_root + 'annotations/imagenet_vid_val.json',
    metric='bbox')
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import Tuple, Union

import torch
import torch.nn as nn
from torch import Tensor

from mmtrack.registry import MODELS


@MODELS.register_module()
class KLMCLoss(nn.Module):
    """KL-divergence loss for probabilistic regression.

    It is computed using Monte Carlo (MC) samples from an arbitrary
    distribution.

    Args:
        eps (float, optional): Defaults to 0.0.
    """

    def __init__(self, eps: float = 0.0):
        super().__init__()
        self.eps = eps

    def forward(self,
                scores: Tensor,
                sample_density: Tensor,
                gt_density: Tensor,
                mc_dim: int = -1) -> Tensor:
        """
        Args:
            scores (Tensor): predicted score values. It has shape
                (num_imgs, num_samples).
            sample_density (Tensor): probability density of the sample
                distribution. It has shape (num_imgs, num_samples).
            gt_density (Tensor): probability density of the ground truth
                distribution. It has shape (num_imgs, num_samples).
            mc_dim (int): dimension of the MC samples.

        Returns:
            torch.Tensor: Calculated loss
        """
        exp_val = scores - torch.log(sample_density + self.eps)

        loss = torch.logsumexp(
            exp_val, dim=mc_dim) - math.log(scores.shape[mc_dim]) - torch.mean(
                scores * (gt_density / (sample_density + self.eps)),
                dim=mc_dim)

        return loss.mean()


@MODELS.register_module()
class KLGridLoss(nn.Module):
    """KL-divergence loss for probabilistic regression.

    It is computed using the grid integration strategy.
    """

    def forward(self,
                scores: Tensor,
                gt_density: Tensor,
                grid_dim: Union[Tuple, int] = -1,
                grid_scale: float = 1.0) -> Tensor:
        """
        Args:
            scores (Tensor): predicted score values. It has shape
                (num_imgs_per_seq, bs, score_map_size, score_map_size).
            gt_density (Tensor): probability density of the ground truth
                distribution. It has shape
                (num_imgs_per_seq, bs, score_map_size, score_map_size).
            grid_dim (int): dimension(s) of the grid.
            grid_scale (float): area of one grid cell.

        Returns:
            torch.Tensor: Calculated loss
        """
        score_corr = grid_scale * torch.sum(scores * gt_density, dim=grid_dim)

        loss = torch.logsumexp(
            scores, dim=grid_dim) + math.log(grid_scale) - score_corr

        return loss.mean()



# Copyright (c) OpenMMLab. All rights reserved.
from .kl_loss import KLGridLoss, KLMCLoss
from .l2_loss import L2Loss
from .multipos_cross_entropy_loss import MultiPosCrossEntropyLoss
from .triplet_loss import TripletLoss

__all__ = [
    'TripletLoss', 'MultiPosCrossEntropyLoss', 'L2Loss', 'KLGridLoss',
    'KLMCLoss'
]



_base_ = [
    './yolox_x_8xb4-80e_crowdhuman-mot17halftrain'
    '_test-mot17halfval.py'
]

img_scale = (896, 1600)

model = dict(
    data_preprocessor=dict(batch_augments=[
        dict(type='BatchSyncRandomResize', random_size_range=(640, 1152))
    ]))

train_pipeline = [
    dict(
        type='Mosaic',
        img_scale=img_scale,
        pad_val=114.0,
        bbox_clip_border=True),
    dict(
        type='RandomAffine',
        scaling_ratio_range=(0.1, 2),
        border=(-img_scale[0] // 2, -img_scale[1] // 2),
        bbox_clip_border=True),
    dict(
        type='MixUp',
        img_scale=img_scale,
        ratio_range=(0.8, 1.6),
        pad_val=114.0,
        bbox_clip_border=True),
    dict(type='YOLOXHSVRandomAug'),
    dict(type='RandomFlip', prob=0.5),
    dict(
        type='Resize',
        scale=img_scale,
        keep_ratio=True,
        clip_object_border=True),
    dict(type='Pad', size_divisor=32, pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(type='FilterAnnotations', min_gt_bbox_wh=(1, 1), keep_empty=False),
    dict(type='PackDetInputs')
]
test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='Resize', scale=img_scale, keep_ratio=True),
    dict(type='Pad', size_divisor=32, pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(
        type='PackDetInputs',
        meta_keys=('img_id', 'img_path', 'ori_shape', 'img_shape',
                   'scale_factor'))
]

train_dataloader = dict(
    dataset=dict(
        type='MultiImageMixDataset',
        _scope_='mmdet',
        dataset=dict(
            type='ConcatDataset',
            _scope_='mmdet',
            datasets=[
                dict(
                    type='CocoDataset',
                    data_root='data/MOT20',
                    ann_file='annotations/train_cocoformat.json',
                    data_prefix=dict(img='train'),
                    filter_cfg=dict(filter_empty_gt=True, min_size=32),
                    metainfo=dict(CLASSES=('pedestrian', )),
                    pipeline=[
                        dict(type='LoadImageFromFile'),
                        dict(type='LoadAnnotations'),
                    ]),
                dict(
                    type='CocoDataset',
                    data_root='data/crowdhuman',
                    ann_file='annotations/crowdhuman_train.json',
                    data_prefix=dict(img='train'),
                    filter_cfg=dict(filter_empty_gt=True, min_size=32),
                    metainfo=dict(CLASSES=('pedestrian', )),
                    pipeline=[
                        dict(type='LoadImageFromFile'),
                        dict(type='LoadAnnotations'),
                    ]),
                dict(
                    type='CocoDataset',
                    data_root='data/crowdhuman',
                    ann_file='annotations/crowdhuman_val.json',
                    data_prefix=dict(img='val'),
                    filter_cfg=dict(filter_empty_gt=True, min_size=32),
                    metainfo=dict(CLASSES=('pedestrian', )),
                    pipeline=[
                        dict(type='LoadImageFromFile'),
                        dict(type='LoadAnnotations'),
                    ]),
            ]),
        pipeline=train_pipeline))
val_dataloader = dict(
    dataset=dict(
        type='CocoDataset',
        data_root='data/MOT17',
        _scope_='mmdet',
        ann_file='annotations/train_cocoformat.json',
        data_prefix=dict(img='train'),
        metainfo=dict(CLASSES=('pedestrian', )),
        test_mode=True,
        pipeline=test_pipeline))
test_dataloader = dict(
    dataset=dict(
        type='CocoDataset',
        data_root='data/MOT20',
        _scope_='mmdet',
        ann_file='annotations/test_cocoformat.json',
        data_prefix=dict(img='test'),
        metainfo=dict(CLASSES=('pedestrian', )),
        test_mode=True,
        pipeline=test_pipeline))

val_evaluator = dict(
    type='mmdet.CocoMetric',
    ann_file='data/MOT17/annotations/train_cocoformat.json',
    metric='bbox',
    format_only=False)
test_evaluator = dict(
    type='mmdet.CocoMetric',
    ann_file='data/MOT20/annotations/test_cocoformat.json',
    metric='bbox',
    format_only=True)



# dataset settings
train_pipeline = [
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[
            dict(type='LoadImageFromFile'),
            dict(
                type='LoadTrackAnnotations',
                with_instance_id=True,
                with_mask=True,
                with_bbox=True),
            dict(type='mmdet.Resize', scale=(640, 360), keep_ratio=True),
            dict(type='mmdet.RandomFlip', prob=0.5),
        ]),
    dict(type='PackTrackInputs', ref_prefix='ref', num_key_frames=1)
]

test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(
        type='LoadTrackAnnotations',
        with_instance_id=True,
        with_mask=True,
        with_bbox=True),
    dict(type='mmdet.Resize', scale=(640, 360), keep_ratio=True),
    dict(type='PackTrackInputs', pack_single_img=True)
]

dataset_type = 'YouTubeVISDataset'
data_root = 'data/youtube_vis_2019/'
dataset_version = data_root[-5:-1]  # 2019 or 2021
# dataloader
train_dataloader = dict(
    batch_size=1,
    num_workers=2,
    persistent_workers=True,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=True),
    batch_sampler=dict(type='mmdet.AspectRatioBatchSampler'),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        dataset_version=dataset_version,
        ann_file='annotations/youtube_vis_2019_train.json',
        data_prefix=dict(img_path='train/JPEGImages'),
        pipeline=train_pipeline,
        load_as_video=True,
        ref_img_sampler=dict(
            num_ref_imgs=1,
            frame_range=100,
            filter_key_img=True,
            method='uniform')))
val_dataloader = dict(
    batch_size=1,
    num_workers=2,
    persistent_workers=True,
    drop_last=False,
    sampler=dict(type='VideoSampler'),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        dataset_version=dataset_version,
        ann_file='annotations/youtube_vis_2019_valid.json',
        data_prefix=dict(img_path='valid/JPEGImages'),
        pipeline=test_pipeline,
        load_as_video=True,
        ref_img_sampler=None,
        test_mode=True))
test_dataloader = val_dataloader



# Copyright (c) OpenMMLab. All rights reserved.
from .base_tracker import BaseTracker
from .byte_tracker import ByteTracker
from .masktrack_rcnn_tracker import MaskTrackRCNNTracker
from .quasi_dense_tao_tracker import QuasiDenseTAOTracker
from .quasi_dense_tracker import QuasiDenseTracker
from .sort_tracker import SORTTracker
from .strongsort_tracker import StrongSORTTracker
from .tracktor_tracker import TracktorTracker

__all__ = [
    'BaseTracker', 'ByteTracker', 'MaskTrackRCNNTracker', 'SORTTracker',
    'QuasiDenseTracker', 'QuasiDenseTAOTracker', 'TracktorTracker',
    'StrongSORTTracker'
]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional, Sequence, Tuple, Union

import cv2
import numpy as np
from mmcv.transforms import BaseTransform, to_tensor
from mmdet.structures.mask import BitmapMasks
from mmengine.structures import InstanceData

from mmtrack.registry import TRANSFORMS
from mmtrack.structures import ReIDDataSample, TrackDataSample


@TRANSFORMS.register_module()
class PackTrackInputs(BaseTransform):
    """Pack the inputs data for the video object detection / multi object
    tracking / single object tracking / video instance segmentation.

    For each value (``List`` type) in the input dict, we concat the first
    `num_key_frames` elements to the first dict with a new key, and the rest
    of elements are concated to the second dict with a new key.
    All the information of images are packed to ``inputs``.
    All the information except images are packed to ``data_samples``.

    Args:
        ref_prefix (str): The prefix of key added to the 'reference' frames.
            Defaults to 'ref'.
        num_key_frames (int): The number of key frames. Defaults to 1.
        num_template_frames (optional, int): The number of template frames. It
            is only used for training in SOT.
        pack_single_img (bool, optional): Whether to only pack single image. If
            True, pack the data as a list additionally. Defaults to False.
        meta_keys (Sequence[str]): Meta keys to be collected in
            ``data_sample.metainfo``. Defaults to None.
        default_meta_keys (tuple): Default meta keys. Defaults to ('img_id',
            'img_path', 'ori_shape', 'img_shape', 'scale_factor',
            'flip', 'flip_direction', 'frame_id', 'is_video_data',
            'video_id', 'video_length', 'instances').
    """
    mapping_table = {
        'gt_bboxes': 'bboxes',
        'gt_bboxes_labels': 'labels',
        'gt_masks': 'masks',
        'gt_instances_id': 'instances_id'
    }

    def __init__(self,
                 ref_prefix: str = 'ref',
                 num_key_frames: int = 1,
                 num_template_frames: Optional[int] = None,
                 pack_single_img: Optional[bool] = False,
                 meta_keys: Optional[dict] = None,
                 default_meta_keys: tuple = ('img_id', 'img_path', 'ori_shape',
                                             'img_shape', 'scale_factor',
                                             'flip', 'flip_direction',
                                             'frame_id', 'is_video_data',
                                             'video_id', 'video_length',
                                             'instances', 'num_left_ref_imgs',
                                             'frame_stride')):
        self.ref_prefix = ref_prefix
        # If ``num_template_frames`` is not None, this class is used in SOT.
        # In this case, we assign the value of ``num_template_frames`` to
        # ``self.num_key_frames`` for the consistency in the processing.
        self.num_key_frames = num_key_frames if num_template_frames is None \
            else num_template_frames
        self.meta_keys = default_meta_keys
        if meta_keys is not None:
            if isinstance(meta_keys, str):
                meta_keys = (meta_keys, )
            else:
                assert isinstance(meta_keys, tuple), \
                    'meta_keys must be str or tuple'
            self.meta_keys += meta_keys

        self.pack_single_img = pack_single_img

    def _cat_same_type_data(
            self,
            data: Union[List, int],
            return_ndarray: bool = True,
            axis: int = 0,
            stack: bool = False) -> Tuple[np.ndarray, np.ndarray]:
        """Concatenate data with the same type.

        Args:
            data (Union[List, int]): Input data.
            return_ndarray (bool, optional): Whether to return ``np.ndarray``.
                Defaults to True.
            axis (int, optional): The axis that concatenating along. Defaults
                to 0.
            stack (bool, optional): Whether to stack all the data. If not,
                using the concatenating operation. Defaults to False.

        Returns:
            Tuple[np.ndarray, np.ndarray]: The first element is the
                concatenated data of key frames, and the second element is the
                concatenated data of reference frames.
        """
        if self.pack_single_img:
            data = [data]
        key_data = data[:self.num_key_frames]
        ref_data = data[self.num_key_frames:] if len(
            data) > self.num_key_frames else None

        if return_ndarray:
            if stack:
                key_data = np.stack(key_data, axis=axis)
                if ref_data is not None:
                    ref_data = np.stack(ref_data, axis=axis)
            else:
                key_data = np.concatenate(key_data, axis=axis)
                if ref_data is not None:
                    ref_data = np.concatenate(ref_data, axis=axis)

        return key_data, ref_data

    def _get_img_idx_map(self, anns: List) -> Tuple[np.ndarray, np.ndarray]:
        """Get the index of images for the annotations. The multiple instances
        in one image need to be denoted the image index when concatenating
        multiple images.

        Args:
            anns (List): Input annotations.

        Returns:
            Tuple[np.ndarray, np.ndarray]: The first element is the
                concatenated indexes of key frames, and the second element is
                the concatenated indexes of reference frames.
        """
        if self.pack_single_img:
            anns = [anns]
        key_img_idx_map = []
        for img_idx, ann in enumerate(anns[:self.num_key_frames]):
            key_img_idx_map.extend([img_idx] * len(ann))
        key_img_idx_map = np.array(key_img_idx_map, dtype=np.int32)
        if len(anns) > self.num_key_frames:
            ref_img_idx_map = []
            for img_idx, ann in enumerate(anns[self.num_key_frames:]):
                ref_img_idx_map.extend([img_idx] * len(ann))
            ref_img_idx_map = np.array(ref_img_idx_map, dtype=np.int32)
        else:
            ref_img_idx_map = None
        return key_img_idx_map, ref_img_idx_map

    def transform(self, results: dict) -> dict:
        """Method to pack the input data.

        Args:
            results (dict): Result dict from the data pipeline.

        Returns:
            dict:
            - 'inputs' (dict[Tensor]): The forward data of models.
            - 'data_samples' (obj:`TrackDataSample`): The annotation info of
                the samples.
        """
        packed_results = dict()
        packed_results['inputs'] = dict()

        # 1. Pack images
        if 'img' in results:
            imgs = results['img']
            key_imgs, ref_imgs = self._cat_same_type_data(imgs, stack=True)
            key_imgs = key_imgs.transpose(0, 3, 1, 2)
            packed_results['inputs']['img'] = to_tensor(key_imgs)

            if ref_imgs is not None:
                ref_imgs = ref_imgs.transpose(0, 3, 1, 2)
                packed_results['inputs'][f'{self.ref_prefix}_img'] = to_tensor(
                    ref_imgs)

        data_sample = TrackDataSample()

        # 2. Pack InstanceData
        if 'gt_ignore_flags' in results:
            gt_ignore_flags = results['gt_ignore_flags']
            (key_gt_ignore_flags,
             ref_gt_ignore_flags) = self._cat_same_type_data(gt_ignore_flags)
            key_valid_idx = key_gt_ignore_flags == 0
            if ref_gt_ignore_flags is not None:
                ref_valid_idx = ref_gt_ignore_flags == 0

        instance_data = InstanceData()
        ignore_instance_data = InstanceData()
        ref_instance_data = InstanceData()
        ref_ignore_instance_data = InstanceData()

        # Flag that whether have recorded the image index
        img_idx_map_flag = False
        for key in self.mapping_table.keys():
            if key not in results:
                continue
            if key == 'gt_masks':
                gt_masks = results[key]
                gt_masks_ndarray = [
                    mask.to_ndarray() for mask in gt_masks
                ] if isinstance(gt_masks, list) else gt_masks.to_ndarray()
                key_gt_masks, ref_gt_masks = self._cat_same_type_data(
                    gt_masks_ndarray)

                mapped_key = self.mapping_table[key]
                if 'gt_ignore_flags' in results:
                    instance_data[mapped_key] = BitmapMasks(
                        key_gt_masks[key_valid_idx], *key_gt_masks.shape[-2:])
                    ignore_instance_data[mapped_key] = BitmapMasks(
                        key_gt_masks[~key_valid_idx], *key_gt_masks.shape[-2:])

                    if ref_gt_masks is not None:
                        ref_instance_data[mapped_key] = BitmapMasks(
                            ref_gt_masks[ref_valid_idx],
                            *key_gt_masks.shape[-2:])
                        ref_ignore_instance_data[mapped_key] = BitmapMasks(
                            ref_gt_masks[~ref_valid_idx],
                            *key_gt_masks.shape[-2:])
                else:
                    instance_data[mapped_key] = BitmapMasks(
                        key_gt_masks, *key_gt_masks.shape[-2:])
                    if ref_gt_masks is not None:
                        ref_instance_data[mapped_key] = BitmapMasks(
                            ref_gt_masks, *ref_gt_masks.shape[-2:])

            else:
                anns = results[key]
                key_anns, ref_anns = self._cat_same_type_data(anns)

                if not img_idx_map_flag:
                    # The multiple instances in one image need to be
                    # denoted the image index when concatenating multiple
                    # images.
                    key_img_idx_map, ref_img_idx_map = self._get_img_idx_map(
                        anns)
                    img_idx_map_flag = True

                mapped_key = self.mapping_table[key]
                if 'gt_ignore_flags' in results:
                    instance_data[mapped_key] = to_tensor(
                        key_anns[key_valid_idx])
                    ignore_instance_data[mapped_key] = to_tensor(
                        key_anns[~key_valid_idx])
                    instance_data['map_instances_to_img_idx'] = to_tensor(
                        key_img_idx_map[key_valid_idx])
                    ignore_instance_data[
                        'map_instances_to_img_idx'] = to_tensor(
                            key_img_idx_map[~key_valid_idx])

                    if ref_anns is not None:
                        ref_instance_data[mapped_key] = to_tensor(
                            ref_anns[ref_valid_idx])
                        ref_ignore_instance_data[mapped_key] = to_tensor(
                            ref_anns[~ref_valid_idx])
                        ref_instance_data[
                            'map_instances_to_img_idx'] = to_tensor(
                                ref_img_idx_map[ref_valid_idx])
                        ref_ignore_instance_data[
                            'map_instances_to_img_idx'] = to_tensor(
                                ref_img_idx_map[~ref_valid_idx])
                else:
                    instance_data[mapped_key] = to_tensor(key_anns)
                    instance_data['map_instances_to_img_idx'] = to_tensor(
                        key_img_idx_map)
                    if ref_anns is not None:
                        ref_instance_data[mapped_key] = to_tensor(ref_anns)
                        ref_instance_data[
                            'map_instances_to_img_idx'] = to_tensor(
                                ref_img_idx_map)

        data_sample.gt_instances = instance_data
        data_sample.ignored_instances = ignore_instance_data
        setattr(data_sample, f'{self.ref_prefix}_gt_instances',
                ref_instance_data)
        setattr(data_sample, f'{self.ref_prefix}_ignored_instances',
                ref_ignore_instance_data)

        # 3. Pack metainfo
        new_img_metas = {}
        for key in self.meta_keys:
            if key not in results:
                continue
            img_metas = results[key]
            key_img_metas, ref_img_metas = self._cat_same_type_data(
                img_metas, return_ndarray=False)
            # To compatible the interface of ``MMDet``, we don't use
            # the fotmat of list when the length of meta information is
            # equal to 1.
            if len(key_img_metas) > 1:
                new_img_metas[key] = key_img_metas
            else:
                new_img_metas[key] = key_img_metas[0]
            if ref_img_metas is not None:
                if len(ref_img_metas) > 1:
                    new_img_metas[f'{self.ref_prefix}_{key}'] = ref_img_metas
                else:
                    new_img_metas[f'{self.ref_prefix}_{key}'] = ref_img_metas[
                        0]

        data_sample.set_metainfo(new_img_metas)

        # 4. Pack some additional properties.
        if 'padding_mask' in results:
            # This property is used in ``STARK`` method in SOT.
            padding_mask = results['padding_mask']
            key_padding_mask, ref_padding_mask = self._cat_same_type_data(
                padding_mask, stack=True)
            data_sample.padding_mask = to_tensor(key_padding_mask)
            if ref_padding_mask is not None:
                setattr(data_sample, f'{self.ref_prefix}_padding_mask',
                        to_tensor(ref_padding_mask))

        packed_results['data_samples'] = data_sample
        return packed_results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'(ref_prefix={self.ref_prefix}, '
        repr_str += f'meta_keys={self.meta_keys}, '
        repr_str += f'default_meta_keys={self.default_meta_keys})'
        return repr_str


@TRANSFORMS.register_module()
class CheckPadMaskValidity(BaseTransform):
    """Check the validity of data. Generally, it's used in such case: The image
    padding masks generated in the image preprocess need to be downsampled, and
    then passed into Transformer model, like DETR. The computation in the
    subsequent Transformer model must make sure that the values of downsampled
    mask are not all zeros.

    Required Keys:

    - gt_bboxes (np.float32)
    - gt_bboxes_labels (np.int32)
    - gt_instances_id (np.int32)
    - gt_masks (BitmapMasks | PolygonMasks)
    - gt_seg_map (np.uint8)
    - gt_ignore_flags (np.bool)
    - img
    - img_shape (optional)
    - jittered_bboxes (optional)
    - padding_mask (np.bool)

    Args:
        stride (int): the max stride of feature map.
    """

    def __init__(self, stride: int):
        self.stride = stride

    def transform(self, results: dict) -> Optional[dict]:
        """The transform function.

        Args:
            results (dict): Result dict contains the data to be checked.

        Returns:
            Optional[dict]: If invalid, return None; otherwise, return original
                input.
        """
        assert 'padding_mask' in results
        masks = results['padding_mask']
        imgs = results['img']
        for mask, img in zip(masks, imgs):
            mask = mask.copy().astype(np.float32)
            img_h, img_w = img.shape[:2]
            feat_h, feat_w = img_h // self.stride, img_w // self.stride
            downsample_mask = cv2.resize(
                mask, dsize=(feat_h, feat_w)).astype(bool)
            if (downsample_mask == 1).all():
                return None
        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'stride={self.stride})'
        return repr_str


@TRANSFORMS.register_module()
class PackReIDInputs(BaseTransform):
    """Pack the inputs data for the ReID.

    The ``meta_info`` item is always populated. The contents of the
    ``meta_info`` dictionary depends on ``meta_keys``. By default
    this includes:

        - ``img_path``: path to the image file.

        - ``ori_shape``: original shape of the image as a tuple (H, W).

        - ``img_shape``: shape of the image input to the network as a tuple
            (H, W). Note that images may be zero padded on the bottom/right
          if the batch tensor is larger than this shape.

        - ``scale``: scale of the image as a tuple (W, H).

        - ``scale_factor``: a float indicating the pre-processing scale.

        -  ``flip``: a boolean indicating if image flip transform was used.

        - ``flip_direction``: the flipping direction.

    Args:
        meta_keys (Sequence[str], optional): The meta keys to saved in the
            ``metainfo`` of the packed ``data_sample``.
    """
    default_meta_keys = ('img_path', 'ori_shape', 'img_shape', 'scale',
                         'scale_factor')

    def __init__(self, meta_keys: Sequence[str] = ()) -> None:
        self.meta_keys = self.default_meta_keys
        if meta_keys is not None:
            if isinstance(meta_keys, str):
                meta_keys = (meta_keys, )
            else:
                assert isinstance(meta_keys, tuple), \
                    'meta_keys must be str or tuple.'
            self.meta_keys += meta_keys

    def transform(self, results: dict) -> dict:
        """Method to pack the input data.

        Args:
            results (dict): Result dict from the data pipeline.

        Returns:
            dict:
            - 'inputs' (dict[Tensor]): The forward data of models.
            - 'data_samples' (obj:`ReIDDataSample`): The meta info of the
                sample.
        """
        packed_results = dict(inputs=dict(), data_samples=None)
        assert 'img' in results, 'Missing the key ``img``.'
        _type = type(results['img'])
        label = results['gt_label']

        if _type == list:
            img = results['img']
            label = np.stack(label, axis=0)  # (N,)
            assert all([type(v) == _type for v in results.values()]), \
                'All items in the results must have the same type.'
        else:
            img = [results['img']]

        img = np.stack(img, axis=3)  # (H, W, C, N)
        img = img.transpose(3, 2, 0, 1)  # (N, C, H, W)
        img = np.ascontiguousarray(img)

        packed_results['inputs'] = to_tensor(img)

        data_sample = ReIDDataSample()
        data_sample.set_gt_label(label)

        meta_info = dict()
        for key in self.meta_keys:
            meta_info[key] = results[key]
        data_sample.set_metainfo(meta_info)
        packed_results['data_samples'] = data_sample

        return packed_results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'(meta_keys={self.meta_keys})'
        return repr_str



_base_ = [
    '../../_base_/models/faster-rcnn_r50-dc5.py',
    '../../_base_/datasets/imagenet_vid_dff_style.py',
    '../../_base_/default_runtime.py'
]
model = dict(
    type='DFF',
    detector=dict(
        train_cfg=dict(
            rpn_proposal=dict(max_per_img=1000),
            rcnn=dict(sampler=dict(num=512)))),
    motion=dict(
        type='FlowNetSimple',
        img_scale_factor=0.5,
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/pretrained_weights/flownet_simple.pth'  # noqa: E501
        )),
    train_cfg=None,
    test_cfg=dict(key_frame_interval=10))

# training schedule
train_cfg = dict(type='EpochBasedTrainLoop', max_epochs=7, val_interval=7)
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')

# learning rate
param_scheduler = [
    dict(
        type='LinearLR',
        start_factor=1.0 / 3,
        by_epoch=False,
        begin=0,
        end=500),
    dict(
        type='MultiStepLR',
        begin=0,
        end=7,
        by_epoch=True,
        milestones=[2, 5],
        gamma=0.1)
]

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(type='SGD', lr=0.01, momentum=0.9, weight_decay=0.0001),
    clip_grad=dict(max_norm=35, norm_type=2))

visualizer = dict(type='DetLocalVisualizer')



# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import Tuple

import torch
import torch.nn as nn
from mmcv.cnn.bricks import ConvModule
from mmcv.ops import PrRoIPool
from mmdet.structures.bbox import bbox_cxcywh_to_xyxy
from mmengine.model import BaseModule
from torch import Tensor

from mmtrack.registry import MODELS
from mmtrack.structures.bbox import (bbox_cxcywh_to_x1y1wh,
                                     bbox_rel_cxcywh_to_xywh,
                                     bbox_xywh_to_rel_cxcywh,
                                     bbox_xyxy_to_x1y1wh)
from mmtrack.utils import OptConfigType, SampleList


class LinearBlock(nn.Module):
    """The linear block. The full pipeline: FC > BN > ReLU.

    Args:
        in_planes (int): The dim of input.
        out_planes (int): The dim of output.
        input_size (int): The size of input.
        bias (bool, optional): Whether to have bias in linear layer. Defaults
            to True.
        batch_norm (bool, optional): Whether to have BN after linear layer.
            Defaults to True.
        relu (bool, optional):  Whether to have ReLU at last. Defaults to True.
    """

    def __init__(self,
                 in_planes: int,
                 out_planes: int,
                 input_size: int,
                 bias: bool = True,
                 batch_norm: bool = True,
                 relu: bool = True):
        super().__init__()
        self.linear = nn.Linear(
            in_planes * input_size * input_size, out_planes, bias=bias)
        self.bn = nn.BatchNorm2d(out_planes) if batch_norm else None
        self.relu = nn.ReLU(inplace=True) if relu else None

    def forward(self, x):
        x = self.linear(x.reshape(x.shape[0], -1))
        if self.bn is not None:
            x = self.bn(x.reshape(x.shape[0], x.shape[1], 1, 1))
        if self.relu is not None:
            x = self.relu(x)
        return x.reshape(x.shape[0], -1)


@MODELS.register_module()
class IouNetHead(BaseModule):
    """Module for IoU prediction.

    Refer to the ATOM paper for a detailed illustration of the architecture.
    `ATOM: <https://arxiv.org/abs/1811.07628>`_.

    Args:
        in_dim (tuple(int), optional): Feature dimensionality from the two
            input backbone layers. Defaults to (128, 256).
        pred_in_dim (tuple(int), optional): Input dimensionality of the
            prediction network. Defaults to (256, 256).
        pred_inter_dim (tuple(int), optional): Intermediate dimensionality in
            the prediction network. Defaults to (256, 256).
        bbox_cfg (dict, optional): The configuration of bbox refinement.
            Defaults to None. Defaults to None.
        train_cfg (dict, optional): The configuration of training.
            Defaults to None.
        loss_bbox (dict, optional): The configuration of loss.
            Defaults to None.
    """

    def __init__(self,
                 in_dim: Tuple = (128, 256),
                 pred_in_dim: Tuple = (256, 256),
                 pred_inter_dim: Tuple = (256, 256),
                 bbox_cfg: OptConfigType = None,
                 train_cfg: OptConfigType = None,
                 test_cfg: OptConfigType = None,
                 loss_bbox: OptConfigType = None,
                 **kwargs):
        super().__init__()
        self.bbox_cfg = bbox_cfg
        self.test_cfg = test_cfg
        self.train_cfg = train_cfg

        self.loss_bbox = MODELS.build(loss_bbox)

        def conv_module(in_planes, out_planes, kernel_size=3, padding=1):
            # The module's pipeline: Conv -> BN -> ReLU.
            return ConvModule(
                in_channels=in_planes,
                out_channels=out_planes,
                kernel_size=kernel_size,
                padding=padding,
                bias=False,
                norm_cfg=dict(type='BN', requires_grad=True),
                act_cfg=dict(type='ReLU'),
                inplace=True)

        # `*_temp` denotes template branch, `*_search` denotes search branch
        # The `number` in the names of variables are block indexes in the
        # backbone or indexes of head layer.
        self.conv3_temp = conv_module(in_dim[0], 128)
        self.roi3_temp = PrRoIPool(3, 1 / 8)
        self.fc3_temp = conv_module(128, 256, padding=0)
        self.fc34_3_temp = conv_module(
            256 + 256, pred_in_dim[0], kernel_size=1, padding=0)

        self.conv4_temp = conv_module(in_dim[1], 256)
        self.roi4_temp = PrRoIPool(1, 1 / 16)
        self.fc34_4_temp = conv_module(
            256 + 256, pred_in_dim[1], kernel_size=1, padding=0)

        self.conv3_search = nn.Sequential(
            conv_module(in_dim[0], 256), conv_module(256, pred_in_dim[0]))
        self.roi3_search = PrRoIPool(5, 1 / 8)
        self.fc3_search = LinearBlock(pred_in_dim[0], pred_inter_dim[0], 5)

        self.conv4_search = nn.Sequential(
            conv_module(in_dim[1], 256), conv_module(256, pred_in_dim[1]))
        self.roi4_search = PrRoIPool(3, 1 / 16)
        self.fc4_search = LinearBlock(pred_in_dim[1], pred_inter_dim[1], 3)

        self.iou_predictor = nn.Linear(
            pred_inter_dim[0] + pred_inter_dim[1], 1, bias=True)

    def init_weights(self):
        """Initialize the parameters of this module."""
        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(
                    m, nn.ConvTranspose2d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight.data, mode='fan_in')
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.uniform_()
                m.bias.data.zero_()

    def predict_iou(self, modulations: Tuple[Tensor], feats: Tensor,
                    proposals: Tensor) -> Tensor:
        """Predicts IoU for the give proposals.

        Args:
            modulations (Tuple(Tensor)): contains the features from two layers.
                The inner tensors are of shape (bs, c, 1, 1)
            feats (Tuple(Tensor)):  IoU features for test images.
                The inner tensors are of shape (bs, c, h, w).
            proposals (Tuple[Tensor]):  Proposal boxes for which the IoU will
                be predicted (bs, num_proposals, 4).

        Returns:
            Tensor: IoU between the proposals with the groundtruth boxes. It's
                of shape (bs, num_proposals).
        """
        # `*_temp` denotes template branch, `*_search` denotes search branch
        # The `number` in the names of variables are block indexes in the
        # backbone or indexes of head layer.
        fc34_3_temp, fc34_4_temp = modulations
        conv3_search, conv4_search = feats
        batch_size = conv3_search.shape[0]

        # Modulation
        conv3_search_att = conv3_search * fc34_3_temp
        conv4_search_att = conv4_search * fc34_4_temp

        # Push the different rois for the same image along the batch dimension
        num_proposals_per_batch = proposals.shape[1]

        # input proposals2 is in format xywh, convert it to x0y0x1y1 format
        proposals_xyxy = torch.cat(
            (proposals[..., 0:2], proposals[..., 0:2] + proposals[..., 2:4]),
            dim=2)

        # Add batch_index to rois
        batch_index = torch.arange(
            batch_size, dtype=torch.float32).reshape(-1, 1,
                                                     1).to(proposals.device)
        roi = torch.cat((batch_index.expand(-1, num_proposals_per_batch,
                                            -1), proposals_xyxy),
                        dim=2)
        roi = roi.reshape(-1, 5)

        roi3_search = self.roi3_search(conv3_search_att, roi)
        roi4_search = self.roi4_search(conv4_search_att, roi)

        fc3_search = self.fc3_search(roi3_search)
        fc4_search = self.fc4_search(roi4_search)

        fc34_search_cat = torch.cat((fc3_search, fc4_search), dim=1)

        iou_pred = self.iou_predictor(fc34_search_cat).reshape(
            batch_size, num_proposals_per_batch)

        return iou_pred

    def get_modulation(self, feats: Tuple[Tensor],
                       bboxes: Tensor) -> Tuple[Tensor, Tensor]:
        """Get modulation vectors for the targets in the search branch.

        Args:
            feats (tuple(Tensor)): Backbone features from template branch.
                It's of shape (bs, c, h, w).
            bboxes (Tensor): Target boxes (x1, y1, x2, y2) in image coords in
                the template branch. It's of shape (bs, 4).

        Returns:
            fc34_3_temp (Tensor): of shape (bs, c, 1, 1).
            fc34_4_temp (Tensor): of shape (bs, c, 1, 1).
        """

        # Add batch_index to rois
        batch_size = bboxes.shape[0]
        batch_index = torch.arange(
            batch_size, dtype=torch.float32).reshape(-1, 1).to(bboxes.device)
        roi = torch.cat((batch_index, bboxes), dim=1)

        # Perform conv and prpool on the feature maps from the backbone
        # `*_temp` denotes template branch, `*_search` denotes search branch
        # The `number` in the names of variables are block indexes in the
        # backbone or indexes of head layer.
        feat3_temp, feat4_temp = feats
        conv3_temp = self.conv3_temp(feat3_temp)
        roi3_temp = self.roi3_temp(conv3_temp, roi)
        fc3_temp = self.fc3_temp(roi3_temp)

        c4_temp = self.conv4_temp(feat4_temp)
        roi4_temp = self.roi4_temp(c4_temp, roi)

        # Concatenate from block 3 and 4
        fc34_temp = torch.cat((fc3_temp, roi4_temp), dim=1)

        fc34_3_temp = self.fc34_3_temp(fc34_temp)
        fc34_4_temp = self.fc34_4_temp(fc34_temp)

        return fc34_3_temp, fc34_4_temp

    def forward(self, feats: Tuple[Tensor]) -> Tuple[Tensor, Tensor]:
        """Get IoU prediction features from a 4 or 5 dimensional backbone
        input.

        Args:
            feats (tuple(Tensor)): Containing the features from backbone with
                shape (bs, c, h, w)

        Returns:
            conv3_search (Tensor): Features from the `conv3_search` branch.
            conv4_search (Tensor): Features from the `conv4_search` branch.
        """
        # `*_temp` denotes template branch, `*_search` denotes search branch
        # The `number` in the names of variables are block indexes in the
        # backbone or indexes of head layer.
        feats = [
            f.reshape(-1, *f.shape[-3:]) if f.dim() == 5 else f for f in feats
        ]
        feat3_search, feat4_search = feats
        conv3_search = self.conv3_search(feat3_search)
        conv4_search = self.conv4_search(feat4_search)

        return conv3_search, conv4_search

    def init_iou_net(self, iou_backbone_feats: Tensor, bboxes: Tensor):
        """Initialize the IoUNet with feature are from the 'layer2' and
        'layer3' of backbone.

        Args:
            iou_backbone_feats (tuple(Tensor)): The features from the backbone.
            bboxes (Tensor): of shape (4, ) or (1, 4) in [cx, cy, w, h] format.
        """
        bboxes = bbox_cxcywh_to_xyxy(bboxes.view(-1, 4))
        # Get modulation vector
        self.iou_modulation = self.get_modulation(iou_backbone_feats, bboxes)

    def optimize_bboxes(self, iou_features: Tuple[Tensor],
                        init_bboxes: Tensor) -> Tuple[Tensor, Tensor]:
        """Optimize the bboxes.

        Args:
            iou_features (tuple(Tensor)): The features used to predict IoU.
            init_bboxes (Tensor): The initialized bboxes with shape (N,4) in
                [cx, cy, w, h] format.

        Returns:
            Tensor: The optimized bboxes with shape (N,4)  in [x, y, w, h]
                format.
            Tensor: The predict IoU of the optimized bboxes with shape (N, ).
        """
        step_length = self.bbox_cfg['box_refine_step_length']
        if isinstance(step_length, (tuple, list)):
            step_length = torch.Tensor([
                step_length[0], step_length[0], step_length[1], step_length[1]
            ]).to(iou_features[0].device).view(1, 1, 4)

        # TODO: simplify this series of transform
        output_bboxes = bbox_cxcywh_to_x1y1wh(init_bboxes)
        output_bboxes = output_bboxes.view(1, -1, 4)
        bboxes_sz_norm = output_bboxes[:, :1, 2:].clone()
        output_bboxes_rel = bbox_xywh_to_rel_cxcywh(output_bboxes,
                                                    bboxes_sz_norm)

        with torch.set_grad_enabled(True):
            for _ in range(self.bbox_cfg['box_refine_iter']):
                # Forward
                bboxes_init_rel = output_bboxes_rel.clone().detach()
                bboxes_init_rel.requires_grad = True

                bboxes_init = bbox_rel_cxcywh_to_xywh(bboxes_init_rel,
                                                      bboxes_sz_norm)
                iou_outputs = self.predict_iou(self.iou_modulation,
                                               iou_features, bboxes_init)
                # Backward
                iou_outputs.backward(gradient=torch.ones_like(iou_outputs))

                # Update bboxes
                output_bboxes_rel = bboxes_init_rel + (
                    step_length * bboxes_init_rel.grad)
                output_bboxes_rel.detach_()

                step_length *= self.bbox_cfg['box_refine_step_decay']

        output_bboxes = bbox_rel_cxcywh_to_xywh(output_bboxes_rel,
                                                bboxes_sz_norm)

        return output_bboxes.view(-1, 4), iou_outputs.detach().view(-1)

    def predict(self, backbone_feats: Tensor, data_samples: SampleList,
                init_bbox: Tensor, sample_center: Tensor,
                scale_factor: float) -> Tensor:
        """Refine the target bounding box.

        Args:
            init_bbox (Tensor): of shape (4, ) or (1, 4) in [cx, cy, w, h]
                formmat.
            backbone_feats (tuple(Tensor)): of shape (1, c, h, w)
            sample_center (Tensor): The center of the cropped
                sample on the original image. It's in [x, y] format.
            scale_factor (float): The size ratio of the cropped patch to the
                resized image.

        Returns:
            Tensor: The refined target bbox in [cx, cy, w, h] format.
        """
        init_bbox = init_bbox.squeeze()
        sample_center = sample_center.squeeze()
        assert sample_center.dim() == 1

        iou_features = self(backbone_feats)
        return self.predict_by_feat(iou_features, init_bbox, sample_center,
                                    scale_factor)

    def predict_by_feat(self, iou_features: Tensor, init_bbox: Tensor,
                        sample_center: Tensor, scale_factor: Tensor) -> Tensor:
        """Refine the target bounding box.

        Args:
            init_bbox (Tensor): The init target bbox.
            iou_features (Tensor): The features for IoU prefiction.
            sample_center (Tensor): The coordinate of the sample center based
                on the original image.
            scale_factor (float): The size ratio of the cropped patch to the
                resized image.

        Returns:
            Tensor: The refined target bbox in [cx, cy, w, h] format.
        """

        # Generate some random initial boxes based on the `init_bbox`
        init_bbox = init_bbox.view(1, 4)
        init_bboxes = init_bbox.clone()
        if self.bbox_cfg['num_init_random_boxes'] > 0:
            square_box_sz = init_bbox[0, 2:].prod().sqrt().item()
            rand_factor = square_box_sz * torch.cat([
                self.bbox_cfg['box_jitter_pos'] * torch.ones(2),
                self.bbox_cfg['box_jitter_sz'] * torch.ones(2)
            ])
            min_edge_size = init_bbox[0, 2:].min() / 3
            rand_bboxes_jitter = (torch.rand(
                self.bbox_cfg['num_init_random_boxes'], 4) - 0.5) * rand_factor
            rand_bboxes_jitter = rand_bboxes_jitter.to(init_bbox.device)
            new_size = (init_bbox[:, 2:] +
                        rand_bboxes_jitter[:, 2:]).clamp(min_edge_size)
            new_center = init_bbox[:, :2] + rand_bboxes_jitter[:, :2]
            init_bboxes = torch.cat([new_center, new_size], 1)
            init_bboxes = torch.cat([init_bbox, init_bboxes])

        # Optimize the boxes
        out_bboxes, out_iou = self.optimize_bboxes(iou_features, init_bboxes)

        return self._bbox_post_process(out_bboxes, out_iou, sample_center,
                                       scale_factor)

    def _bbox_post_process(self, out_bboxes: Tensor, out_ious: Tensor,
                           sample_center: Tensor,
                           scale_factor: float) -> Tensor:
        """The post process about bbox.

        Args:
            out_bboxes (Tensor): The several optimized bboxes.
            out_ious (Tensor): The IoUs about the optimized bboxes.
            sample_center (Tensor): The coordinate of the sample center based
                on the original image.
            scale_factor (float): The size ratio of the cropped patch to the
                resized image.

        Returns:
            Tensor: The refined target bbox in [cx, cy, w, h] format.
        """
        # Remove weird boxes according to the ratio of aspect
        out_bboxes[:, 2:].clamp_(1)
        aspect_ratio = out_bboxes[:, 2] / out_bboxes[:, 3]
        keep_ind = (aspect_ratio < self.bbox_cfg['max_aspect_ratio']) * (
            aspect_ratio > 1 / self.bbox_cfg['max_aspect_ratio'])
        out_bboxes = out_bboxes[keep_ind, :]
        out_ious = out_ious[keep_ind]

        # If no box found
        if out_bboxes.shape[0] == 0:
            return None

        # Predict box
        k = self.bbox_cfg['iounet_topk']
        topk = min(k, out_bboxes.shape[0])
        _, inds = torch.topk(out_ious, topk)
        # in [x,y,w,h] format
        predicted_box = out_bboxes[inds, :].mean(0)

        # Convert the bbox of the cropped sample to that of original image.
        # TODO: this postprocess about mapping back can be moved to other place
        new_bbox_center = predicted_box[:2] + predicted_box[2:] / 2
        new_bbox_center = (new_bbox_center - self.test_cfg['img_sample_size'] /
                           2) * scale_factor + sample_center
        new_target_size = predicted_box[2:] * scale_factor

        return torch.cat([new_bbox_center, new_target_size], dim=-1)

    def _gauss_density_centered(self, x: Tensor, sigma: Tensor) -> Tensor:
        """Evaluate the probability density of a Gaussian centered at zero.

        args:
            x (Tensor): of (num_smples, 4) shape.
            sigma (Tensor): Standard deviations with (1, 4, 2) shape.
        """

        return torch.exp(-0.5 * (x / sigma)**2) / (
            math.sqrt(2 * math.pi) * sigma)

    def _gmm_density_centered(self, x: Tensor, sigma: Tensor) -> Tensor:
        """Evaluate the probability density of a GMM centered at zero.

        args:
            x(Tensor): of (num_smples, 4) shape.
            sigma (Tensor): Tensor of standard deviations with (1, 4, 2) shape.
        """
        if x.dim() == sigma.dim() - 1:
            x = x[..., None]
        elif not (x.dim() == sigma.dim() and x.shape[-1] == 1):
            raise ValueError('Last dimension must be the gmm sigmas.')

        # ``product`` along feature dim of ``bbox```, ``mean``` along component
        # dim of ``sigma``.
        return self._gauss_density_centered(x, sigma).prod(-2).mean(-1)

    def _sample_gmm_centered(self,
                             sigma: Tensor,
                             num_samples: int = 1) -> Tuple[Tensor, Tensor]:
        """Sample from a GMM distribution centered at zero.

        Args:
            sigma (Tensor): Standard deviations of bbox coordinates with
                [4, 2] shape.
            num_samples (int, optional): The number of samples.

        Returns:
            x_centered (Tensor): of shape (num_samples, num_dims)
            prob_density (Tensor): of shape (num_samples, )
        """
        num_components = sigma.shape[-1]
        num_dims = sigma.shape[-2]

        sigma = sigma.reshape(1, num_dims, num_components)

        # Sampling component index
        k = torch.randint(
            num_components, size=(num_samples, ), dtype=torch.int64)
        sigma_samples = sigma[0, :, k].t()

        x_centered = sigma_samples * torch.randn(num_samples, num_dims).to(
            sigma_samples.device)
        prob_density = self._gmm_density_centered(x_centered, sigma)

        return x_centered, prob_density

    def get_targets(self, bbox: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """Generate the training targets for search images.

        Args:
            bbox (Tensor): The bbox of (N, 4) shape in [x, y, w, h] format.

        Returns:
            Tuple[Tensor, Tensor, Tensor]:
                ``proposals``: proposals with [num_samples, 4] shape.
                ``proposal_density``: proposal density with [num_samples, ]
                    shape.
                ``gt_density``: groundtruth density with [num_samples, ] shape.
        """
        bbox = bbox.clone().reshape(-1, 4)
        bbox_wh = bbox[:, 2:]

        if not hasattr(self, 'proposals_sigma'):
            center_sigma = torch.tensor(
                [s[0] for s in self.train_cfg['proposals_sigma']])
            size_sigma = torch.tensor(
                [s[1] for s in self.train_cfg['proposals_sigma']])
            # of shape (4, len(train_cfg['proposal_sigma']))
            self.proposals_sigma = torch.stack(
                (center_sigma, center_sigma, size_sigma, size_sigma),
                dim=0).to(bbox.device)

        if not hasattr(self, 'gt_bboxes_sigma'):
            # of shape (1, 4)
            self.gt_bboxes_sigma = torch.tensor(
                (self.train_cfg['gt_bboxes_sigma'][0],
                 self.train_cfg['gt_bboxes_sigma'][0],
                 self.train_cfg['gt_bboxes_sigma'][1],
                 self.train_cfg['gt_bboxes_sigma'][1]),
                dtype=torch.float32,
                device=bbox.device).reshape(-1, 4)

        # Sample boxes
        proposals_rel_centered, proposal_density = self._sample_gmm_centered(
            self.proposals_sigma,
            self.train_cfg['num_samples'] * bbox.shape[0])

        # Add mean and map back
        # of shape (num_seq*bs, 4)
        mean_box_rel = bbox_xywh_to_rel_cxcywh(bbox, bbox_wh)
        # the first num_samples elements along zero dim are
        # from the same image
        # of shape (num_samples*bbox.shape[0] , bbox.shape[-1])
        mean_box_rel = mean_box_rel.unsqueeze(1).expand(
            -1, self.train_cfg['num_samples'],
            -1).reshape(-1, mean_box_rel.shape[-1])
        proposals_rel = proposals_rel_centered + mean_box_rel
        # of shape (num_samples * bbox.shape[0], bbox.shape[-1])
        bbox_wh = bbox_wh.unsqueeze(1).expand(-1,
                                              self.train_cfg['num_samples'],
                                              -1).reshape(
                                                  -1, bbox_wh.shape[-1])
        proposals = bbox_rel_cxcywh_to_xywh(proposals_rel, bbox_wh)

        # of shape (num_samples, )
        gt_density = self._gauss_density_centered(
            proposals_rel_centered, self.gt_bboxes_sigma).prod(-1)

        if self.train_cfg['add_first_bbox']:
            proposals = torch.cat((bbox, proposals))
            proposal_density = torch.cat(
                (torch.tensor([-1.]), proposal_density))
            gt_density = torch.cat((torch.tensor([1.]), gt_density))

        return proposals, proposal_density, gt_density

    def loss(self, template_feats: Tuple[Tensor], search_feats: Tuple[Tensor],
             batch_data_samples: SampleList, **kwargs) -> dict:
        """Perform forward propagation and loss calculation of the tracking
        head on the features of the upstream network.

        Args:
            template_feats (tuple[Tensor, ...]): Tuple of Tensor with
                shape (N, C, H, W) denoting the multi level feature maps of
                exemplar images.
            search_feats (tuple[Tensor, ...]): Tuple of Tensor with shape
                (N, C, H, W) denoting the multi level feature maps of
                search images.
            batch_data_samples (List[:obj:`TrackDataSample`]): The Data
                Samples. It usually includes information such as `gt_instance`.

        Returns:
            dict: A dictionary of loss components.
        """
        batch_size = len(batch_data_samples)
        batch_gt_bboxes = []
        batch_img_metas = []
        batch_search_gt_bboxes = []

        for data_sample in batch_data_samples:
            batch_img_metas.append(data_sample.metainfo)
            batch_gt_bboxes.append(data_sample.gt_instances['bboxes'])
            batch_search_gt_bboxes.append(
                data_sample.search_gt_instances['bboxes'])

        # Extract the first train sample in each sequence
        template_feats = [feat[:batch_size, ...] for feat in template_feats]
        batch_gt_bboxes = torch.stack(batch_gt_bboxes, dim=1)

        # Get modulation vector
        modulations = self.get_modulation(template_feats, batch_gt_bboxes[0])

        iou_feats = self(search_feats)
        num_search_imgs_per_seq = batch_data_samples[0].search_gt_instances[
            'bboxes'].shape[0]
        # (num_seq*bs, c). The first `bs` tensors along
        # zero-dim are from different images of a batch.
        modulations = [
            feat.repeat(num_search_imgs_per_seq, 1, 1, 1)
            for feat in modulations
        ]

        return self.loss_by_feat(modulations, iou_feats, batch_gt_bboxes,
                                 batch_search_gt_bboxes)

    def loss_by_feat(self, modulations: Tuple[Tensor],
                     iou_feats: Tuple[Tensor], batch_gt_bboxes: Tensor,
                     batch_search_gt_bboxes: Tensor) -> dict:
        """Compute loss.

        Args:
            modulations (Tuple[Tensor]): The modulation features.
            iou_feats (Tuple[Tensor]): The features for iou prediction.
            batch_gt_bboxes (Tensor): The gt_bboxes in a batch.
            batch_search_gt_bboxes (Tensor): The search gt_bboxes in a batch.

        Returns:
            dict: a dictionary of loss components.
        """
        batch_search_gt_bboxes = torch.stack(
            batch_search_gt_bboxes, dim=1).view(-1, 4)
        batch_search_gt_bboxes_xywh = bbox_xyxy_to_x1y1wh(
            batch_search_gt_bboxes)
        (proposals, proposals_density, search_gt_bboxes_density
         ) = self.get_targets(batch_search_gt_bboxes_xywh)

        proposals = proposals.view(-1, self.train_cfg['num_samples'],
                                   proposals.shape[-1])
        proposals_density = proposals_density.view(
            -1, self.train_cfg['num_samples'])
        search_gt_bboxes_density = search_gt_bboxes_density.view(
            -1, self.train_cfg['num_samples'])
        pred_iou = self.predict_iou(modulations, iou_feats, proposals)

        loss_bbox = self.loss_bbox(
            pred_iou,
            sample_density=proposals_density,
            gt_density=search_gt_bboxes_density,
            mc_dim=1) * self.train_cfg['loss_weights']['bbox']

        return dict(loss_bbox=loss_bbox)



# Copyright (c) OpenMMLab. All rights reserved.
import math
from copy import deepcopy
from typing import List, Optional, Tuple, Union

import torch
import torch.nn.functional as F
from mmdet.structures.bbox.transforms import bbox_xyxy_to_cxcywh
from torch import Tensor
from torch.nn.modules.batchnorm import _BatchNorm
from torch.nn.modules.conv import _ConvNd

from mmtrack.registry import MODELS
from mmtrack.structures import TrackDataSample
from mmtrack.utils import (InstanceList, OptConfigType, OptMultiConfig,
                           SampleList)
from .base import BaseSingleObjectTracker


@MODELS.register_module()
class Stark(BaseSingleObjectTracker):
    """STARK: Learning Spatio-Temporal Transformer for Visual Tracking.

    This single object tracker is the implementation of `STARk
    <https://arxiv.org/abs/2103.17154>`_.

    Args:
        backbone (dict): the configuration of backbone network.
        neck (dict, optional): the configuration of neck network.
            Defaults to None.
        head (dict, optional): the configuration of head network.
            Defaults to None.
        init_cfg (dict, optional): the configuration of initialization.
            Defaults to None.
        frozen_modules (str | list | tuple, optional): the names of frozen
            modules. Defaults to None.
        train_cfg (dict, optional): the configuratioin of train.
            Defaults to None.
        test_cfg (dict, optional): the configuration of test.
            Defaults to None.
    """

    def __init__(self,
                 backbone: dict,
                 neck: Optional[dict] = None,
                 head: Optional[dict] = None,
                 pretrains: Optional[dict] = None,
                 train_cfg: Optional[dict] = None,
                 test_cfg: Optional[dict] = None,
                 frozen_modules: Optional[Union[List[str], Tuple[str],
                                                str]] = None,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptMultiConfig = None):
        super(Stark, self).__init__(data_preprocessor, init_cfg)
        head.update(test_cfg=test_cfg)
        self.backbone = MODELS.build(backbone)
        self.neck = MODELS.build(neck)
        self.head = MODELS.build(head)

        self.test_cfg = test_cfg
        self.train_cfg = train_cfg

        self.num_templates = self.test_cfg['num_templates']

        # Set the update interval
        self.update_intervals = self.test_cfg.get('update_intervals', None)
        if isinstance(self.update_intervals, (int, float)):
            self.update_intervals = [int(self.update_intervals)
                                     ] * self.num_templates

        if frozen_modules is not None:
            self.freeze_module(frozen_modules)

    def init_weights(self):
        """Initialize the weights of modules in single object tracker."""
        # We don't use the `init_weights()` function in BaseModule, since it
        # doesn't support the initialization method from `reset_parameters()`
        # in Pytorch.
        if self.with_backbone:
            self.backbone.init_weights()

        if self.with_neck:
            for m in self.neck.modules():
                if isinstance(m, _ConvNd) or isinstance(m, _BatchNorm):
                    m.reset_parameters()

        if self.with_head:
            self.head.init_weights()

    def extract_feat(self, img: Tensor) -> Tensor:
        """Extract the features of the input image.

        Args:
            img (Tensor): image of shape (N, C, H, W).

        Returns:
            tuple(Tensor): the multi-level feature maps, and each of them is
                    of shape (N, C, H // stride, W // stride).
        """
        feat = self.backbone(img)
        feat = self.neck(feat)
        return feat

    def get_cropped_img(self, img: Tensor, target_bbox: Tensor,
                        search_area_factor: float,
                        output_size: float) -> Union[Tensor, float, Tensor]:
        """ Crop Image
        Only used during testing
        This function mainly contains two steps:
        1. Crop `img` based on target_bbox and search_area_factor. If the
        cropped image/mask is out of boundary of `img`, use 0 to pad.
        2. Resize the cropped image/mask to `output_size`.

        args:
            img (Tensor): of shape (1, C, H, W)
            target_bbox (Tensor): in [cx, cy, w, h] format
            search_area_factor (float): Ratio of crop size to target size.
            output_size (float): the size of output cropped image
                (always square).
        returns:
            img_crop_padded (Tensor): of shape (1, C, output_size, output_size)
            resize_factor (float): the ratio of original image scale to cropped
                image scale.
            pdding_mask (Tensor): the padding mask caused by cropping. It's
                of shape (1, output_size, output_size).
        """
        cx, cy, w, h = target_bbox.split((1, 1, 1, 1), dim=-1)

        img_h, img_w = img.shape[2:]
        # 1. Crop image
        # 1.1 calculate crop size and pad size
        crop_size = math.ceil(math.sqrt(w * h) * search_area_factor)
        if crop_size < 1:
            raise Exception('Too small bounding box.')

        x1 = torch.round(cx - crop_size * 0.5).long()
        x2 = x1 + crop_size
        y1 = torch.round(cy - crop_size * 0.5).long()
        y2 = y1 + crop_size

        x1_pad = max(0, -x1)
        x2_pad = max(x2 - img_w + 1, 0)
        y1_pad = max(0, -y1)
        y2_pad = max(y2 - img_h + 1, 0)

        # 1.2 crop image
        img_crop = img[..., y1 + y1_pad:y2 - y2_pad, x1 + x1_pad:x2 - x2_pad]

        # 1.3 pad image
        img_crop_padded = F.pad(
            img_crop,
            pad=(x1_pad, x2_pad, y1_pad, y2_pad),
            mode='constant',
            value=0)
        # 1.4 generate padding mask
        _, _, img_h, img_w = img_crop_padded.shape
        end_x = None if x2_pad == 0 else -x2_pad
        end_y = None if y2_pad == 0 else -y2_pad
        padding_mask = torch.ones((img_h, img_w),
                                  dtype=torch.float32,
                                  device=img.device)
        padding_mask[y1_pad:end_y, x1_pad:end_x] = 0.

        # 2. Resize cropped image and padding mask
        resize_factor = output_size / crop_size
        img_crop_padded = F.interpolate(
            img_crop_padded, (output_size, output_size),
            mode='bilinear',
            align_corners=False)

        padding_mask = F.interpolate(
            padding_mask[None, None], (output_size, output_size),
            mode='bilinear',
            align_corners=False).squeeze(dim=0).type(torch.bool)

        return img_crop_padded, resize_factor, padding_mask

    def init(self, img: Tensor):
        """Initialize the single object tracker in the first frame.

        Args:
            img (Tensor): input image of shape (1, C, H, W).
        """
        self.memo.z_dict_list = []  # store templates
        # get the 1st template
        z_patch, _, z_mask = self.get_cropped_img(
            img, self.memo.bbox, self.test_cfg['template_factor'],
            self.test_cfg['template_size']
        )  # z_patch of shape [1,C,H,W];  z_mask of shape [1,H,W]

        z_feat = self.extract_feat(z_patch)

        self.z_dict = dict(feat=z_feat, mask=z_mask)
        self.memo.z_dict_list.append(self.z_dict)

        # get other templates
        for _ in range(self.num_templates - 1):
            self.memo.z_dict_list.append(deepcopy(self.z_dict))

    def update_template(self, img: Tensor, bbox: Union[List, Tensor],
                        conf_score: float):
        """Update the dymanic templates.

        Args:
            img (Tensor): of shape (1, C, H, W).
            bbox (list | Tensor): in [cx, cy, w, h] format.
            conf_score (float): the confidence score of the predicted bbox.
        """
        for i, update_interval in enumerate(self.update_intervals):
            if self.memo.frame_id % update_interval == 0 and conf_score > 0.5:
                z_patch, _, z_mask = self.get_cropped_img(
                    img,
                    bbox,
                    self.test_cfg['template_factor'],
                    output_size=self.test_cfg['template_size'])
                z_feat = self.extract_feat(z_patch)
                # the 1st element of z_dict_list is the template from the 1st
                # frame
                self.memo.z_dict_list[i + 1] = dict(feat=z_feat, mask=z_mask)

    def track(self, img: Tensor, data_samples: SampleList) -> InstanceList:
        """Track the box of previous frame to current frame `img`.

        Args:
            img (Tensor): of shape (1, C, H, W).
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as ``gt_instances`` and 'metainfo'.

        Returns:
            InstanceList: Tracking results of each image after the postprocess.
                - scores: a Tensor denoting the score of best_bbox.
                - bboxes: a Tensor of shape (4, ) in [x1, x2, y1, y2]
                format, and denotes the best tracked bbox in current frame.
        """
        # get the search patches
        x_patch, resize_factor, x_mask = self.get_cropped_img(
            img, self.memo.bbox, self.test_cfg['search_factor'],
            self.test_cfg['search_size']
        )  # bbox: of shape (x1, y1, w, h), x_mask: of shape (1, h, w)

        x_feat = self.extract_feat(x_patch)
        x_dict = dict(feat=x_feat, mask=x_mask)
        head_inputs = self.memo.z_dict_list + [x_dict]
        results = self.head.predict(head_inputs, data_samples, self.memo.bbox,
                                    resize_factor)

        if results[0].scores.item() != -1:
            # get confidence score (whether the search region is reliable)
            crop_bbox = bbox_xyxy_to_cxcywh(results[0].bboxes.squeeze())
            self.update_template(img, crop_bbox, results[0].scores.item())

        return results

    def predict_vot(self, inputs: dict, data_samples: List[TrackDataSample]):
        raise NotImplementedError(
            'STARK does not support testing on VOT datasets yet.')

    def loss(self, inputs: dict, data_samples: List[TrackDataSample],
             **kwargs) -> dict:
        """Forward of training.

        Args:
            inputs (dict[Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size. The T denotes the number of
                key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.

            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.

        Return:
            dict: A dictionary of loss components.
        """
        template_padding_mask = [
            data_sample.padding_mask for data_sample in data_samples
        ]
        template_padding_mask = torch.stack(template_padding_mask, dim=0)
        search_padding_mask = [
            data_sample.search_padding_mask for data_sample in data_samples
        ]
        search_padding_mask = torch.stack(search_padding_mask, dim=0)

        search_img = inputs['search_img']
        assert search_img.dim(
        ) == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        template_img = inputs['img']
        assert template_img.dim(
        ) == 5, 'The img must be 5D Tensor (N, T, C, H, W).'

        head_inputs = []
        for i in range(self.num_templates):
            z_feat = self.extract_feat(template_img[:, i])
            z_dict = dict(feat=z_feat, mask=template_padding_mask[:, i])
            head_inputs.append(z_dict)
        x_feat = self.extract_feat(search_img[:, 0])
        x_dict = dict(feat=x_feat, mask=search_padding_mask[:, 0])
        head_inputs.append(x_dict)

        losses = self.head.loss(head_inputs, data_samples)

        return losses



# Copyright (c) OpenMMLab. All rights reserved.
from .siamrpn_backbone_unfreeze_hook import SiamRPNBackboneUnfreezeHook
from .visualization_hook import TrackVisualizationHook
from .yolox_mode_switch_hook import YOLOXModeSwitchHook

__all__ = [
    'YOLOXModeSwitchHook', 'TrackVisualizationHook',
    'SiamRPNBackboneUnfreezeHook'
]



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from typing import Dict, List, Tuple, Union

import addict
from mmdet.structures.bbox import bbox_cxcywh_to_xyxy, bbox_xyxy_to_cxcywh
from mmengine.model import BaseModel
from mmengine.structures import InstanceData
from torch import Tensor

from mmtrack.evaluation import bbox2region
from mmtrack.structures.bbox import calculate_region_overlap, quad2bbox_cxcywh
from mmtrack.utils import (ForwardResults, InstanceList, OptConfigType,
                           OptMultiConfig, OptSampleList, SampleList)


class BaseSingleObjectTracker(BaseModel, metaclass=ABCMeta):
    """Base class for single object tracker.

    Args:
        data_preprocessor (dict or ConfigDict, optional): The pre-process
           config of :class:`TrackDataPreprocessor`.  it usually includes,
            ``pad_size_divisor``, ``pad_value``, ``mean`` and ``std``.
        init_cfg (dict or list[dict]): Initialization config dict.
    """

    def __init__(self,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptMultiConfig = None) -> None:
        super().__init__(
            data_preprocessor=data_preprocessor, init_cfg=init_cfg)

    def freeze_module(self, module: Union[List[str], Tuple[str], str]) -> None:
        """Freeze module during training."""
        if isinstance(module, str):
            modules = [module]
        else:
            if not (isinstance(module, list) or isinstance(module, tuple)):
                raise TypeError('module must be a str or a list.')
            else:
                modules = module
        for module in modules:
            m = getattr(self, module)
            m.eval()
            for param in m.parameters():
                param.requires_grad = False

    @property
    def with_backbone(self):
        """bool: whether the framework has a backbone"""
        return hasattr(self, 'backbone') and self.backbone is not None

    @property
    def with_neck(self):
        """bool: whether the framework has a neck"""
        return hasattr(self, 'neck') and self.neck is not None

    @property
    def with_head(self):
        """bool: whether the framework has a head"""
        return hasattr(self, 'head') and self.head is not None

    def forward(self,
                inputs: Dict[str, Tensor],
                data_samples: OptSampleList = None,
                mode: str = 'predict',
                **kwargs) -> ForwardResults:
        """The unified entry for a forward process in both training and test.

        The method should accept three modes: "tensor", "predict" and "loss":

        - "tensor": Forward the whole network and return tensor or tuple of
        tensor without any post-processing, same as a common nn.Module.
        - "predict": Forward and return the predictions, which are fully
        processed to a list of :obj:`TrackDataSample`.
        - "loss": Forward and return a dict of losses according to the given
        inputs and data samples.

        Note that this method doesn't handle neither back propagation nor
        optimizer updating, which are done in the :meth:`train_step`.

        Args:
            inputs (Dict[str, Tensor]): of shape (N, T, C, H, W)
                encoding input images. Typically these should be mean centered
                and std scaled. The N denotes batch size. The T denotes the
                number of key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
            data_samples (list[:obj:`TrackDataSample`], optional): The
                annotation data of every samples. Defaults to None.
            mode (str): Return what kind of value. Defaults to 'predict'.

        Returns:
            The return type depends on ``mode``.

            - If ``mode="tensor"``, return a tensor or a tuple of tensor.
            - If ``mode="predict"``, return a list of :obj:`TrackDataSample`.
            - If ``mode="loss"``, return a dict of tensor.
        """
        if mode == 'loss':
            return self.loss(inputs, data_samples, **kwargs)
        elif mode == 'predict':
            return self.predict(inputs, data_samples, **kwargs)
        elif mode == 'tensor':
            return self._forward(inputs, data_samples, **kwargs)
        else:
            raise RuntimeError(f'Invalid mode "{mode}". '
                               'Only supports loss, predict and tensor mode')

    @abstractmethod
    def loss(self, inputs: Dict[str, Tensor], data_samples: SampleList,
             **kwargs) -> Union[dict, tuple]:
        """Calculate losses from a batch of inputs and data samples."""
        pass

    def predict(self, inputs: Dict[str, Tensor], data_samples: SampleList,
                **kwargs) -> SampleList:
        """Predict results from a batch of inputs and data samples with post-
        processing.

        Args:
            inputs (dict[Tensor]): of shape (N, T, C, H, W)
                encodingbinput images. Typically these should be mean centered
                and stdbscaled. The N denotes batch size. The T denotes the
                number of key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
                In test mode, T = 1 and there is only ``img`` and no
                ``ref_img``.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as ``gt_instances`` and ``metainfo``.

        Returns:
            list[obj:`TrackDataSample`]: Tracking results of the
                input images. Each TrackDataSample usually contains
                ``pred_det_instances`` or ``pred_track_instances``.
        """
        test_mode = self.test_cfg.get('test_mode', 'OPE')
        assert test_mode in ['OPE', 'VOT']
        if test_mode == 'VOT':
            pred_track_instances = self.predict_vot(inputs, data_samples)
        else:
            pred_track_instances = self.predict_ope(inputs, data_samples)
        track_data_samples = deepcopy(data_samples)
        for _data_sample, _pred_instances in zip(track_data_samples,
                                                 pred_track_instances):
            _data_sample.pred_track_instances = _pred_instances
        return track_data_samples

    def predict_ope(self, inputs: dict, data_samples: SampleList):

        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(0) == 1, \
            'Only support 1 batch size per gpu in test mode'
        img = img[0]

        metainfo = data_samples[0].metainfo
        frame_id = metainfo.get('frame_id', -1)
        assert frame_id >= 0

        if frame_id == 0:
            # We store the current frame id, previous bbox and template
            # information in ``self.memo``.
            self.memo = addict.Dict()
            self.memo.frame_id = frame_id
            gt_bboxes = data_samples[0].gt_instances['bboxes']
            self.memo.bbox = quad2bbox_cxcywh(gt_bboxes)
            self.init(img)
            results = [InstanceData()]
            results[0].bboxes = bbox_cxcywh_to_xyxy(self.memo.bbox)[None]
            results[0].scores = gt_bboxes.new_tensor([-1.])
        else:
            self.memo.frame_id = frame_id
            results = self.track(img, data_samples)
            self.memo.bbox = bbox_xyxy_to_cxcywh(results[0].bboxes.squeeze())

        return results

    def predict_vot(self, inputs: dict,
                    data_samples: SampleList) -> List[InstanceData]:
        """Test using VOT test mode.

        Args:
            inputs (Dict[str, Tensor]): of shape (N, T, C, H, W)
                encoding input images. Typically these should be mean centered
                and std scaled. The N denotes batch size. The T denotes the
                number of key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
                In test mode, T = 1 and there is only ``img`` and no
                ``ref_img``.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as ``gt_instances`` and 'metainfo'.

        Returns:
            List[:obj:`InstanceData`]: Object tracking results of each image
            after the post process. Each item usually contains following keys.
                - scores (Tensor): Classification scores, has a shape (1, )
                - bboxes (Tensor): Has a shape (1, 4),
                  the last dimension 4 arrange as [x1, y1, x2, y2].
        """
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(0) == 1, \
            'Only support 1 batch size per gpu in test mode'
        img = img[0]

        gt_bboxes = data_samples[0].gt_instances['bboxes']
        metainfo = data_samples[0].metainfo
        frame_id = metainfo.get('frame_id', -1)
        assert frame_id >= 0

        if frame_id == 0:
            self.init_frame_id = 0
        if self.init_frame_id == frame_id:
            # initialization
            # We store the previous bbox、 current frame id and some template
            # information in ``self.memo``.
            self.memo = addict.Dict()
            self.memo.frame_id = frame_id
            self.memo.bbox = quad2bbox_cxcywh(gt_bboxes)
            self.init(img)
            # 1 denotes the initialization state
            results = [InstanceData()]
            results[0].bboxes = gt_bboxes.new_tensor([[1.]])
            results[0].scores = gt_bboxes.new_tensor([-1.])
        elif self.init_frame_id > frame_id:
            # 0 denotes unknown state, namely the skipping frame after failure
            self.memo.frame_id = frame_id
            results = [InstanceData()]
            results[0].bboxes = gt_bboxes.new_tensor([[0.]])
            results[0].scores = gt_bboxes.new_tensor([-1.])
        else:
            # normal tracking state
            self.memo.frame_id = frame_id
            results = self.track(img, data_samples)
            self.memo.bbox = bbox_xyxy_to_cxcywh(results[0].bboxes.squeeze())

            # convert bbox to region for overlap calculation
            track_bbox = results[0].bboxes[0].cpu().numpy()
            track_region = bbox2region(track_bbox)
            gt_region = bbox2region(gt_bboxes[0].cpu().numpy())

            if 'img_shape' in metainfo:
                image_shape = metainfo['img_shape']
                image_wh = (image_shape[1], image_shape[0])
            else:
                image_wh = None
                Warning('image shape are need when calculating bbox overlap')
            overlap = calculate_region_overlap(
                track_region, gt_region, bounds=image_wh)
            if overlap <= 0:
                # tracking failure
                self.init_frame_id = frame_id + 5
                # 2 denotes the failure state
                results[0].bboxes = img.new_tensor([[2.]])

        return results

    @abstractmethod
    def init(img: Tensor):
        """Initialize the single object tracker in the first frame.

        Args:
            img (Tensor): of shape (1, C, H, W) encoding original input
                image.
        """
        pass

    @abstractmethod
    def track(img: Tensor, data_samples: SampleList) -> InstanceList:
        """Track the box of previous frame to current frame `img`.

        Args:
            img (Tensor): of shape (1, C, H, W) encoding original input
                image.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as ``gt_instances`` and 'metainfo'.

        Returns:
            InstanceList: Tracking results of each image after the postprocess.
                - scores: a Tensor denoting the score of best_bbox.
                - bboxes: a Tensor of shape (4, ) in [x1, x2, y1, y2]
                format, and denotes the best tracked bbox in current frame.
        """
        pass

    def _forward(self,
                 inputs: Dict[str, Tensor],
                 data_samples: OptSampleList = None,
                 **kwargs):
        """Network forward process. Usually includes backbone, neck and head
        forward without any post-processing.

         Args:
            inputs (Dict[str, Tensor]): Inputs with shape (N, C, H, W).
            data_samples (List[:obj:`TrackDataSample`], optional): The
                Data Samples. It usually includes information such as
                `gt_instance`.

        Returns:
            tuple[list]: A tuple of features from ``head`` forward.
        """
        raise NotImplementedError(
            "_forward function (namely 'tensor' mode) is not supported now")



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np
from mmdet.datasets.transforms import LoadAnnotations as MMDet_LoadAnnotations

from mmtrack.registry import TRANSFORMS


@TRANSFORMS.register_module()
class LoadTrackAnnotations(MMDet_LoadAnnotations):
    """Load and process the ``instances`` and ``seg_map`` annotation provided
    by dataset.

    The annotation format is as the following:

    .. code-block:: python

        {
            'instances':
            [
                {
                # List of 4 numbers representing the bounding box of the
                # instance, in (x1, y1, x2, y2) order.
                'bbox': [x1, y1, x2, y2],

                # Label of image classification.
                'bbox_label': 1,

                # Used in tracking.
                # Id of instances.
                'instance_id': 100,

                # Used in instance/panoptic segmentation. The segmentation mask
                # of the instance or the information of segments.
                # 1. If list[list[float]], it represents a list of polygons,
                # one for each connected component of the object. Each
                # list[float] is one simple polygon in the format of
                # [x1, y1, ..., xn, yn] (n≥3). The Xs and Ys are absolute
                # coordinates in unit of pixels.
                # 2. If dict, it represents the per-pixel segmentation mask in
                # COCO’s compressed RLE format. The dict should have keys
                # “size” and “counts”.  Can be loaded by pycocotools
                'mask': list[list[float]] or dict,

                }
            ]
            # Filename of semantic or panoptic segmentation ground truth file.
            'seg_map_path': 'a/b/c'
        }

    After this module, the annotation has been changed to the format below:

    .. code-block:: python

        {
            # In (x1, y1, x2, y2) order, float type. N is the number of bboxes
            # in an image
            'gt_bboxes': np.ndarray(N, 4)
             # In int type.
            'gt_bboxes_labels': np.ndarray(N, )
             # In built-in class
            'gt_masks': PolygonMasks (H, W) or BitmapMasks (H, W)
             # In uint8 type.
            'gt_seg_map': np.ndarray (H, W)
             # in (x, y, v) order, float type.
        }

    Required Keys:

    - height (optional)
    - width (optional)
    - instances

      - bbox (optional)
      - bbox_label
      - instance_id (optional)
      - mask (optional)
      - ignore_flag (optional)

    - seg_map_path (optional)

    Added Keys:

    - gt_bboxes (np.float32)
    - gt_bboxes_labels (np.int32)
    - gt_instances_id (np.int32)
    - gt_masks (BitmapMasks | PolygonMasks)
    - gt_seg_map (np.uint8)
    - gt_ignore_flags (np.bool)

    Args:
        with_instance_id (bool): Whether to parse and load the instance id
            annotation. Defaults to False.
    """

    def __init__(self, with_instance_id: bool = False, **kwargs) -> None:
        super().__init__(**kwargs)
        self.with_instance_id = with_instance_id

    def _load_bboxes(self, results: dict) -> None:
        """Private function to load bounding box annotations.

        The only difference is that we record the type of `gt_ignore_flags`
        as np.int32.

        Args:
            results (dict): Result dict from :obj:``mmcv.BaseDataset``.
        Returns:
            dict: The dict contains loaded bounding box annotations.
        """
        gt_bboxes = []
        gt_ignore_flags = []
        for instance in results['instances']:
            # The datasets which are only format in evaluation don't have
            # groundtruth boxes.
            if 'bbox' in instance:
                gt_bboxes.append(instance['bbox'])
            if 'ignore_flag' in instance:
                gt_ignore_flags.append(instance['ignore_flag'])

        if len(gt_bboxes) != len(gt_ignore_flags):
            # There may be no ``gt_ignore_flags`` in some cases, we treat them
            # as all False in order to keep the length of ``gt_bboxes`` and
            # ``gt_ignore_flags`` the same
            gt_ignore_flags = [False] * len(gt_bboxes)

        if len(gt_bboxes) > 0 and len(gt_bboxes[0]) == 8:
            # The bbox of VOT2018 has (N, 8) shape and it's not possible to be
            # empty.
            results['gt_bboxes'] = np.array(
                gt_bboxes, dtype=np.float32).reshape(-1, 8)
        else:
            # Some tasks, such as VID, may have empty bboxes and their bboxes
            # need to be reshaped to (0, 4) format forcely in order to be
            # compatible with ``TransformBroadcaster``.
            results['gt_bboxes'] = np.array(
                gt_bboxes, dtype=np.float32).reshape(-1, 4)

        results['gt_ignore_flags'] = np.array(gt_ignore_flags, dtype=np.bool)

    def _load_instances_id(self, results: dict) -> None:
        """Private function to load instances id annotations.

        Args:
            results (dict): Result dict from :obj :obj:``mmcv.BaseDataset``.

        Returns:
            dict: The dict contains loaded instances id annotations.
        """
        gt_instances_id = []
        for instance in results['instances']:
            gt_instances_id.append(instance['instance_id'])
        results['gt_instances_id'] = np.array(gt_instances_id, dtype=np.int32)

    def transform(self, results: dict) -> dict:
        """Function to load multiple types annotations.

        Args:
            results (dict): Result dict from :obj:``mmcv.BaseDataset``.

        Returns:
            dict: The dict contains loaded bounding box, label, instances id
            and semantic segmentation and keypoints annotations.
        """
        results = super().transform(results)
        if self.with_instance_id:
            self._load_instances_id(results)
        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'(with_bbox={self.with_bbox}, '
        repr_str += f'with_label={self.with_label}, '
        repr_str += f'with_instance_id={self.with_instance_id}, '
        repr_str += f'with_mask={self.with_mask}, '
        repr_str += f'with_seg={self.with_seg}, '
        repr_str += f'poly2mask={self.poly2mask}, '
        repr_str += f"imdecode_backend='{self.imdecode_backend}', "
        repr_str += f'file_client_args={self.file_client_args})'
        return repr_str



# Copyright (c) OpenMMLab. All rights reserved.
from mmtrack.registry import DATASETS
from .mot_challenge_dataset import MOTChallengeDataset


@DATASETS.register_module()
class DanceTrackDataset(MOTChallengeDataset):
    """Dataset for DanceTrack: https://github.com/DanceTrack/DanceTrack.

    All content is inherited from MOTChallengeDataset.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



# Copyright (c) OpenMMLab. All rights reserved.
from .base_video_metrics import BaseVideoMetric
from .coco_video_metric import CocoVideoMetric
from .mot_challenge_metrics import MOTChallengeMetrics
from .reid_metrics import ReIDMetrics
from .sot_metrics import SOTMetric
from .tao_metrics import TAOMetric
from .youtube_vis_metrics import YouTubeVISMetric

__all__ = [
    'ReIDMetrics', 'BaseVideoMetric', 'CocoVideoMetric', 'YouTubeVISMetric',
    'MOTChallengeMetrics', 'SOTMetric', 'TAOMetric'
]



# dataset settings
dataset_type = 'TaoDataset'
data_root = 'data/tao/'

# data pipeline
train_pipeline = [
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[
            dict(type='LoadImageFromFile'),
            dict(type='LoadTrackAnnotations', with_instance_id=True),
            dict(
                type='RandomChoiceResize',
                scales=[(1333, 640), (1333, 672), (1333, 704), (1333, 736),
                        (1333, 768), (1333, 800)],
                resize_type='mmdet.Resize',
                keep_ratio=True),
            dict(type='mmdet.RandomFlip', prob=0.5),
        ]),
    dict(type='PackTrackInputs', ref_prefix='ref')
]

test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadTrackAnnotations', with_instance_id=True),
    dict(type='mmdet.Resize', scale=(1333, 800), keep_ratio=True),
    dict(
        type='PackTrackInputs',
        pack_single_img=True,
        meta_keys=('frame_index', 'neg_category_ids',
                   'not_exhaustive_category_ids'))
]

# dataloader
train_dataloader = dict(
    batch_size=2,
    num_workers=2,
    persistent_workers=True,
    sampler=dict(type='DefaultSampler', shuffle=True),
    batch_sampler=dict(type='mmdet.AspectRatioBatchSampler'),
    dataset=dict(
        type='ClassBalancedDataset',
        oversample_thr=1e-3,
        dataset=dict(
            type=dataset_type,
            data_root='data/lvis/',
            load_as_video=False,
            ref_img_sampler=dict(num_ref_imgs=1, frame_range=0),
            metainfo=dict(CLASSES=data_root + 'annotations/tao_classes.txt'),
            ann_file='annotations/lvisv0.5+coco_train.json',
            data_prefix=dict(img_path='train'),
            pipeline=train_pipeline)))
val_dataloader = dict(
    batch_size=1,
    num_workers=2,
    persistent_workers=True,
    drop_last=False,
    sampler=dict(type='VideoSampler'),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        metainfo=dict(CLASSES=data_root + 'annotations/tao_classes.txt'),
        ann_file='annotations/validation_482_classes.json',
        ref_img_sampler=None,
        load_as_video=True,
        test_mode=True,
        pipeline=test_pipeline))
test_dataloader = val_dataloader
